'''`EsoRex <http://www.eso.org/sci/software/cpl/esorex.html>`_ is a standard
execution environment for CPL recipes provided by `ESO <http://www.eso.org>`_.
'''

import os
import logging
from .recipe import Recipe
from . import logger

def load_sof(source):
    '''Read an :program:`EsoRex` SOF file.

    :param source: SOF ("Set Of Files") file object or string with SOF
        file content.
    :type source: :class:`str` or :class:`file`

    These files contain the raw and calibration files for a recipe.  The
    content of the file is returned as a map with the tag as key and the list
    of file names as value.

    The result of this function may directly set as :attr:`cpl.Recipe.calib`
    attribute::
    
      import cpl
      myrecipe = cpl.Recipe('muse_bias')
      myrecipe.calib = cpl.esorex.read_sof(open('muse_bias.sof'))
    
    .. note::

      The raw data frame is silently ignored wenn setting
      :attr:`cpl.Recipe.calib` for MUSE recipes. Other recipes ignore the raw
      data frame only if it was set manually as :attr:`cpl.Recipe.tag` or in
      :attr:`cpl.Recipe.tags` since there is no way to automatically
      distinguish between them.

    '''
    if isinstance(source, str):
        return load_sof(open(source) if os.path.exists(source) else source.split('\n'))
    else:
        res = dict()
        for line in source:
            if not line or line.startswith('#'):
                continue
            ls = line.split()
            fn = ls[0]
            key = ls[1]
            if key not in  res:
                res[key] = fn
            elif isinstance(res[key], list):
                res[key].append(fn)
            else:
                res[key] = [ res[key], fn ]
        return res

def load_rc(source = None):
    '''Read an :program:`EsoRex` configuration file.

    :param source: Configuration file object, or string with file content. 
                   If not set, the :program:`EsoRex` config file
                   :file:`~/.esorex/esorex.rc` is used.
    :type source: :class:`str` or :class:`file`

    These files contain configuration parameters for :program:`EsoRex` or
    recipes. The content of the file is returned as a map with the (full)
    parameter name as key and its setting as string value.

    The result of this function may directly set as :attr:`cpl.Recipe.param`
    attribute::
    
      import cpl
      myrecipe = cpl.Recipe('muse_bias')
      myrecipe.param = cpl.esorex.load_rc('muse_bias.rc')

    '''
    if source is None:
        source = open(os.path.expanduser('~/.esorex/esorex.rc'))
    if isinstance(source, str):
        return load_rc(open(source) if os.path.exists(source) else source.split('\n'))
    else:
        res = dict()
        for line in source:
            if not line or not line.strip() or line.startswith('#'):
                continue
            name = line.split('=', 1)[0]
            value = line.split('=', 1)[1]
            if name and value:
                res[name.strip()] = value.strip()
        return res

def init(source = None):
    '''Set up the logging and the recipe search path from the
    :file:`esorex.rc` file.

    :param source: Configuration file object, or string with file content. 
        If not set, the esorex config file :file:`~/.esorex/esorex.rc` is used.
    :type source: :class:`str` or :class:`file`
    '''

    rc = load_rc(source)
    if 'esorex.caller.recipe-dir' in rc:
        Recipe.path = rc['esorex.caller.recipe-dir'].split(':')
    if 'esorex.caller.msg-level' in rc:
        msg.level = rc['esorex.caller.msg-level']
    if 'esorex.caller.log-level' in rc:
        log.level = rc['esorex.caller.log-level']
    if 'esorex.caller.log-dir' in rc:
        log.dir = rc['esorex.caller.log-dir']
    if 'esorex.caller.log-file' in rc:
        log.filename = rc['esorex.caller.log-file']

class CplLogger(object):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARN = logging.WARN
    ERROR = logging.ERROR
    OFF = logging.CRITICAL + 1

    def __init__(self):
        self.handler = None
        self._component = False
        self._time = False
        self._threadid = False
        self.format = None
        self.dir = None
        self._level = CplLogger.OFF

    def _init_handler(self):
        if not self.handler:
            self.handler = logging.StreamHandler()
            logging.getLogger().addHandler(self.handler)
            self.handler.setLevel(self._level)
            self.handler.setFormatter(logging.Formatter(self.format,
                                                        '%H:%M:%S'))

    def _shutdown_handler(self):
        if self.handler:
            logging.getLogger().removeHandler(self.handler)
            self.handler.close()
            self.handler = None

    @property
    def level(self):
        '''Log level for output to the terminal. Any of
        [ DEBUG, INFO, WARN, ERROR, OFF ].
        '''
        return self._level

    @level.setter
    def level(self, level):
        if isinstance(level, (str)):
            level = logger.level[level.upper()]
        if level == CplLogger.OFF:
            self._shutdown_handler()
        else:
            self._init_handler()
            logging.getLogger().setLevel(logging.DEBUG)
            if self.handler:
                self.handler.setLevel(level)
        self._level = level

    @property
    def format(self):
        '''Output format. 

        .. seealso :: `logging.LogRecord attributes <http://docs.python.org/library/logging.html#logrecord-attributes>`_ 

           Key mappings in the logging output.'''
        return self._format

    @format.setter
    def format(self, fmt):
        if fmt is None:
            fmt = '%(asctime)s ' if self._time else ''
            fmt += '[%(levelname)7s]'
            fmt += '[%(threadName)s] ' if self._threadid else ' '
            fmt += '%(name)s: ' if self._component else ''
            fmt += '%(message)s'
        if self.handler:
            self.handler.setFormatter(logging.Formatter(fmt, '%H:%M:%S'))
        self._format = fmt

    @property
    def component(self):
        '''If :obj:`True`, attach the component name to output messages.
        '''
        return self._component

    @component.setter
    def component(self, enable):
        self._component = enable
        self.format = None

    @property
    def time(self):
        '''If :obj:`True`, attach a time tag to output messages.
        '''
        return self._time

    @time.setter
    def time(self, enable):
        self._time = enable
        self.format = None

    @property
    def threadid(self):
        '''If :obj:`True`, attach a thread tag to output messages.
        '''
        return self._threadid

    @threadid.setter
    def threadid(self, enable):
        self._threadid = enable
        self.format = None

class CplFileLogger(CplLogger):
    def __init__(self):
        CplLogger.__init__(self)
        self._filename = None
        self.threadid = True
        self.component = True
        self.time = True
        self.level = CplLogger.INFO

    def _init_handler(self):
        if not self.handler:
            if self._filename:
                if self.dir:
                    fname = os.path.join(self.dir, self._filename)
                    self.handler = logging.FileHandler(fname)
                else:
                    self.handler = logging.FileHandler(self._filename)
            else:
                self.handler = None
            if self.handler:
                logging.getLogger().addHandler(self.handler)
                self.handler.setLevel(self._level)
                self.handler.setFormatter(logging.Formatter(self.format,
                                                            '%H:%M:%S'))

    @property
    def filename(self):
        '''Log file name.
        '''
        return self._filename

    @filename.setter
    def filename(self, name):
        if self._filename != name:
            self._shutdown_handler()
            self._filename = name
            self._init_handler()


msg = CplLogger()
'''This variable is a :class:`CplLogger` instance that provides a convienience
stream handler similar to the terminal logging functionality of the CPL. It
basically does the same as::

  import logging

  log = logging.getLogger()
  log.setLevel(logging.INFO)
  ch = logging.StreamHandler()
  ch.setLevel(logging.OFF)
  ch.setFormatter(logging.Formatter('[%(levelname)7s] %(message)s'))
  log.addHandler(ch)
'''

log = CplFileLogger()
'''This variable is a :class:`CplFileLogger` instance that provides a convienience
file handler similar to the file logging functionality of the CPL. It
basically does the same as::

  import logging

  log = logging.getLogger()
  log.setLevel(logging.INFO)
  ch = logging.FileHandler(filename)
  ch.setLevel(logging.INFO)
  ch.setFormatter(logging.Formatter('%(asctime)s [%(levelname)7s] %(funcName)s: %(message)s'))
  log.addHandler(ch)
'''



import collections
import os
import signal
import logging

from astropy.io import fits

class Result(object):
    def __init__(self, directory, res, input_len = 0, logger = None, 
                 output_format = fits.HDUList):
        '''Build an object containing all result frames.

        Calling :meth:`cpl.Recipe.__call__` returns an object that contains
        all result ('production') frames in attributes. All results for one
        tag are summarized in one attribute of the same name. 

        If the argument `output_format` is :class:`astropy.io.fits.HDUList`
        (default), then the attribute content is either a
        :class:`astropy.io.fits.HDUList` or a class:`list` of HDU lists,
        depending on the recipe and the call: If the recipe produces one out
        put frame of a tag per input file, the attribute contains a list if
        the recipe was called with a list, and if the recipe was called with a
        single input frame, the result attribute will also contain a single
        input frame. If the recipe combines all input frames to one output
        frame, a single :class:`astropy.io.fits.HDUList` es returned,
        independent of the input parameters.

        Similarly, if the argument `output_format` is set to :class:`str`, the
        attribute content is either a :class:`str` or a class:`list` of
        :class:`str`, containing the paths of output files. In this case,
        removing the output files is suppressed.

        .. todo:: This behaviour is made on some heuristics based on the
           number and type of the input frames. The heuristics will go wrong
           if there is only one input frame, specified as a list, but the
           recipe tries to summarize the input. In this case, the attribute
           will contain a list where a single :class:`astropy.io.fits.HDUList`
           was expected. To solve this problem, the "MASTER" flag has to be
           forwarded from the (MUSE) recipe which means that it should be
           exported by the recipe -- this is a major change since it probably
           leads into a replacement of CPLs recipeconfig module by something
           more sophisticated. And this is not usable for non-MUSE recipes
           anyway. So, we will skip this to probably some distant future.
        '''
        self.dir = os.path.abspath(directory)
        logger.join()
        if res[2][0]:
            raise CplError(res[2][0], res[1], logger)
        self.tags = set()
        for tag, frame in res[0]:
            if (output_format == fits.HDUList): 
                # Move the file to the base dir to avoid NFS problems
                outframe = os.path.join(
                    os.path.dirname(self.dir), 
                    '%s.%s' % (os.path.basename(self.dir), frame))
                os.rename(os.path.join(self.dir, frame), outframe)
            else:
                outframe = os.path.join(self.dir, frame)
            if output_format == fits.HDUList:
                hdulist = fits.open(outframe, memmap = True, mode = 'update')
                hdulist.readall()
                os.remove(outframe)
                outframe = hdulist
            tag = tag
            if tag not in self.__dict__:
                self.__dict__[tag] = outframe if input_len != 1 \
                    else [ outframe ]
                self.tags.add(tag)
            elif isinstance(self.__dict__[tag], (fits.HDUList, str)):
                self.__dict__[tag] = [ self.__dict__[tag], outframe ]
            else:
                self.__dict__[tag].append(outframe)
        mtracefname = os.path.join(self.dir, 'recipe.mtrace')
        mtrace = None
        if os.path.exists(mtracefname):
            try:
                mtrace = os.popen("mtrace %s" % mtracefname).read();
            except:
                mtrace = None
        self.stat = Stat(res[2], mtrace)
        self.error = CplError(res[2][0], res[1], logger) if res[1] else None
        self.log = logger.entries if logger else None

    def __getitem__(self, key):
        if key in self.tags:
            return self.__dict__[key]
        else:
            raise KeyError(key)

    def __contains__(self, key):
        return key in self.tags

    def __len__(self):
        return len(self.tags)

    def __iter__(self):
        return iter((key, self.__dict__[key]) for key in self.tags)

class Stat(object):
    def __init__(self, stat, mtrace):
        self.return_code = stat[0]
        self.user_time = stat[1]
        self.sys_time = stat[2]
        self.memory_is_empty = { -1:None, 0:False, 1:True }[stat[3]]
        self.mtrace = mtrace;

class CplError(Exception):
    '''Error message from the recipe.

    If the CPL recipe invocation returns an error, it is converted into a
    :class:`cpl.CplError` exception and no frames are returned. Also, the
    error is notified in the log file.

    The exception is raised on recipe invocation, or when accessing the result
    frames if the recipe was started in background
    (:attr:`cpl.Recipe.threaded` set to :obj:`True`).

    Attributes:

    .. attribute:: code

       The CPL error code returned from the recipe.

    .. attribute:: msg

       The supplied error message. 
    
    .. attribute:: filename

       The source file name where the error occurred.

    .. attribute:: line

       The line number where the error occurred.

    .. attribute:: log

       Log lines of the recipe that lead to this exception.

       .. seealso:: :class:`cpl.logger.LogList`

    .. attribute:: next_error
     
       Next error, or :obj:`None`.

    '''
    def __init__(self, retval, res, logger = None):
        self.retval = retval
        self.log = logger.entries if logger else None
        self.next_error = None
        if not res:
            self.code, self.msg, self.file, self.line, self.function = (
                None, None, None, None, None)
        else:
            self.code, self.msg, self.file, self.line, self.function = res[0]
            o = self
            for r in res[1:]:
                o.next_error = CplError(retval, [ r ], logger)
                o = o.next_error
    
    def __iter__(self):
        class Iter(object):
            current = self
            def __next__(self):
                if Iter.current is None:
                    raise StopIteration
                s = Iter.current
                Iter.current = Iter.current.next_error
                return s
            def next(self):
                return self.__next__()
        return Iter()

    def __str__(self):
        if self.code is None:
            s = 'Unspecified'
        else:
            s = "%s (%i) in %s() (%s:%i)" % (self.msg, self.code, 
                                             self.function, self.file, 
                                             self.line) 
        if self.next_error:
            for e in self.next_error:
                s += "\n    %s (%i) in %s() (%s:%i)" % (e.msg, e.code, 
                                                        e.function, e.file, 
                                                        e.line) 
        return s

class RecipeCrash(Exception):
    '''Recipe crash exception

    If the CPL recipe crashes with a SIGSEV or a SIGBUS, the C stack trace is
    tried to conserved in this exception. The stack trace is obtained with the
    GNU debugger gdb. If the debugger is not available, or if the debugger
    cannot be attached to the crashed recipe, the Exception remains empty.

    When converted to a string, the Exception will return a stack trace
    similar to the Python stack trace.

    The exception is raised on recipe invocation, or when accessing the result
    frames if the recipe was started in background
    (:attr:`cpl.Recipe.threaded` set to :obj:`True`).

    Attributes:

    .. attribute:: elements

       List of stack elements, with the most recent element (the one that
       caused the crash) at the end. Each stack element is a 
       :func:`collections.namedtuple` with the following attributes:

       .. attribute:: filename
 
          Source file name, including full path, if available.

       .. attribute:: line

          Line number, if available

       .. attribute:: func

          Function name, if available

       .. attribute:: params

          Dictionary parameters the function was called with.  The key here is
          the parameter name, the value is a string describing the value set.

       .. attribute:: localvars

          Dictionary of local variables of the function, if available.  The
          key here is the parameter name, the value is a string describing the
          value set.

    .. attribute:: signal

          Signal that caused the crash.
    '''

    StackElement = collections.namedtuple('StackElement', 
                                          'filename line func params localvars')
    signals = {signal.SIGSEGV:'SIGSEV: Segmentation Fault', 
               signal.SIGBUS:'SIGBUS: Bus Error',
               signal.SIGHUP:'SIGHUP: Hangup',
               signal.SIGABRT:'SIGABRT: Abnormal process termination',
               signal.SIGTERM:'SIGTERM: Terminated by user',
               signal.SIGQUIT:'SIGQUIT: Quit',
               signal.SIGFPE:'SIGFPE: Arithmetic Exception',
               signal.SIGINT:'SIGINT: Interrupt (Ctrl-C)',
               None:'Memory inconsistency detected'}
    def __init__(self, bt_file):
        self.elements = []
        current_element = None
        parse_functions = True
        parse_sourcelist = False
        sourcefiles = dict()
        self.signal = None
        self.lines = []
        for line in bt_file:
            self.lines.append(line)
            if line.startswith('Received signal:'):
                self.signal = int(line.split(':')[1])
            if line.startswith('Memory corruption'):
                self.signal = None
            elif line.find('signal handler called') >= 0:
                del self.elements[:]
            elif parse_functions:
                if line.startswith('#'):
                    try:
                        current_element = self._parse_function_line(line)
                    except StopIteration:
                        parse_functions = False
                elif current_element is not None:
                    self._add_variable(current_element.localvars, line)
            if line.startswith('Source files'):
                parse_sourcelist = True
                parse_functions = False
            elif parse_sourcelist:
                sourcefiles.update(dict((os.path.basename(s.strip()), s.strip())
                                        for s in line.split(',') 
                                        if s.rfind('/') > 0 ))
        self.elements = [ RecipeCrash.StackElement(sourcefiles.get(e.filename, 
                                                                   e.filename),
                                                   e.line, e.func, e.params, 
                                                   e.localvars) 
                          for e in self.elements ]
        Exception.__init__(self, str(self))

    def _add_variable(self, variables, line):
        s = line.strip().split('=', 1)
        if len(s) > 1:
            variables[s[0].strip()] = s[1].strip()

    def _parse_function_line(self, line):
        s = line.split()
        funcname = s[3] if s[1].startswith('0x') else s[1]
        if funcname.startswith('Py'):
            raise StopIteration()
        pars = {}
        for fp in line[line.find('(')+1:line.rfind(')')].split(','):
            self._add_variable(pars, fp)
        l = line[line.rfind(')')+1:].split()
        if not l:
            return None
        source = l[-1].split(':')
        filename = source[0]
        lineno = int(source[1]) if len(source) > 1 else None
        current_element = RecipeCrash.StackElement(filename, lineno, 
                                                   funcname, pars, {})
        self.elements.insert(0, current_element)
        return current_element
        
    def log(self, logger):
        '''Put the content of the crash into the log.
        '''
        log = logging.getLogger('%s' % logger.name)
        log.error('Recipe crashed. Traceback (most recent call last):')
        for e in self.elements:
            logc = logging.getLogger('%s.%s' % (logger.name, e.func))
            logc.error('  File "%s", %sin %s\n',e.filename, 
                       'line %i, ' % e.line if e.line else '', 
                       e.func)
            if os.path.exists(e.filename) and e.line:
                logc.error('    %s\n',
                           open(e.filename).readlines()[e.line-1].strip())
            if e.params:
                logc.error('  Parameters:')
                for p, v in e.params.items():
                    logc.error('    %s = %s', p, v)
            if e.localvars:
                logc.error('  Local variables:')
                for p, v in e.localvars.items():
                    logc.error('    %s = %s', p, v)
        log.error(RecipeCrash.signals.get(self.signal, 
                                          '%s: Unknown' % str(self.signal)))

    def __repr__(self):
        return 'RecipeCrash()'

    def __str__(self):
        s = 'Recipe Traceback (most recent call last):\n'
        for e in self.elements:
            s += '  File "%s", %sin %s\n' % ((e.filename), 
                                             'line %i, ' % e.line if e.line 
                                             else '', 
                                             e.func)
            if os.path.exists(e.filename) and e.line:
                s += '    %s\n' % open(e.filename).readlines()[e.line-1].strip()
        s += RecipeCrash.signals.get(self.signal, '%s: Unknown' % str(self.signal))
        return s




from __future__ import absolute_import
import os
from astropy.io import fits

from . import md5sum

class FrameConfig(object):
    '''Frame configuration. 

    Each :class:`FrameConfig` object stores information about one the data
    type a recipe can process. They are used for defining the calibration
    files. However, since this information is not generally provided by CPL
    recipes, it contains only dummy information, except for the MUSE recipes.

    The objects stores a frame tag, a unique identifier for a certain kind of
    frame, the minimum and maximum number of frames needed.

    Attributes:

    .. attribute:: tag 

       Category tag name. The tag name is used to distinguish between
       different types of files. An examples of tag names is 'MASTER_BIAS'
       which specifies the master bias calibration file(s).

    .. attribute:: min

       Minimal number of frames, or :obj:`None` if not specified. A frame is
       required if the :attr:`min` is set to a value greater than 0.

    .. attribute:: max 

       Maximal number of frames, or :obj:`None` if not specified

    .. attribute:: frames

       List of frames (file names or :class:`astropy.io.fits.HDUList` objects)
       that are assigned to this frame type.
    '''
    def __init__(self, tag, min_frames = 0, max_frames = 0, frames = None):
        self.tag = tag
        self.min = min_frames if min_frames > 0 else None
        self.max = max_frames if max_frames > 0 else None
        self.frames = frames
        self.__doc__ = self._doc()

    def extend_range(self, min_frames, max_frames):
        if self.min is not None:
            self.min = min(self.min, min_frames) if min_frames is not None \
                else None
        if self.max is not None:
            self.max = max(self.max, max_frames) if max_frames is not None \
                else None

    def set_range(self, min_frames, max_frames):
        self.min = min_frames
        self.max = max_frames

    def __str__(self):
        return str(self.frames)

    def __repr__(self):
        return 'FrameDef(%s, frames=%s)' % (repr(self.tag), repr(self.frames))

    def _doc(self):
        if self.max is None or self.min is None:
            r = ' one frame or list of frames'
        elif self.max == 1:
            r = ' one frame'
        elif self.min > 1 and self.max > self.min:
            r = ' list of %i-%i frames' % (self.min, self.max)
        elif self.max > 1:
            r = ' one frame or list of max. %i frames' % self.max
        elif self.min > 1:
            r = ' list of min. %i frames' % self.max
        else:
            r = ' one frame or list of frames'
        if not self.min:
            r += ' (optional)'
        return r

    def __getitem__(self, i):
        return (self.tag, self.frames)[i]
    

class FrameList(object):
    def __init__(self, recipe, other = None):
        self._recipe = recipe
        self._values = dict()
        if isinstance(other, self.__class__):
            self._set_items((o.tag, o.frames) for o in other)
        elif isinstance(other, dict):
            self._set_items(other.items())
        elif other:
            self._set_items(other)

    def _set_items(self, l):
        for o in l:
            self[o[0]] = o[1]

    @property
    def _cpl_dict(self):
        cpl_frameconfigs = self._recipe._recipe.frameConfig()
        if cpl_frameconfigs is None:
            return None
        s = dict()
        for configs in cpl_frameconfigs:
            c_cfg = configs[1]
            for f in c_cfg:
                if f[0] in s:
                    s[f[0]].extend_range(f[1], f[2])
                elif f[0] in self._values:
                    s[f[0]] = self._values[f[0]]
                    s[f[0]].set_range(f[1], f[2])
                else:
                    s[f[0]] = FrameConfig(f[0], f[1], f[2])
                    self._values[f[0]] = s[f[0]]
        return s

    @property
    def _dict(self):
        return self._cpl_dict or self._values

    def __iter__(self):
        return iter(self._dict.values())

    def __getitem__(self, key):
        return self._dict[key]

    def __setitem__(self, key, value):
        d = self._cpl_dict
        if d is not None:
            d[key].frames = value
        else:
            self._values.setdefault(key, FrameConfig(key)).frames = value

    def __delitem__(self, key):
        self._dict[key].frames = None

    def __contains__(self, key):
        return key in self._dict

    def __len__(self):
        return len(self._dict)
        
    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        if key.startswith('_'):
            super(FrameList, self).__setattr__(key, value)
        else:
            self[key] = value

    def __delattr__(self, key):
        del self[key]

    def __dir__(self):
        return self._dict.keys()

    def __repr__(self):
        return repr(dict(iter(self)))

    def __str__(self):
        return str(dict(iter(self)))

    def __eq__(self, other):
        return dict(iter(self)) == other

    @property
    def __doc__(self):
        r = 'Frames for recipe %s.\n\nAttributes:\n' % (
            self._recipe.name)
        for s in self:
            r += '%s: %s\n' % (self._key(s), s.__doc__)
        return r        

    def _aslist(self, frames):
        flist = FrameList(self._recipe, self)
        if frames is not None:
            flist._set_items(frames.items())
        return [(f.tag, f.frames) for f in flist]

def mkabspath(frames, tmpdir):
    '''Convert all filenames in the frames list into absolute paths.

    :class:`astropy.io.fits.HDUList`s will be converted to temporary files
    located in the temporary directory tmpdir.

    The replacement is done in-place. The function will return the list of
    temporary files.

    param frames: :class:`list` of (tag, frame) tuples with frame being either
                  a file name or a HDU list.

    param tmpdir: directory where the temporary files are being created.
    '''
    
    tmpfiles = list()
    for i, frame in enumerate(frames):
        if isinstance(frame[1], fits.HDUList):
            md5 = md5sum.update_md5(frame[1])
            filename = os.path.abspath(os.path.join(tmpdir, '%s_%s.fits' 
                                                    % (frame[0], md5[:8])))
            try:
                os.remove(filename)
            except:
                pass
            frames[i] = ( frame[0], filename )
            tmpfiles.append(filename)
            frame[1].writeto(filename)
        else:
            frames[i] = ( frame[0], os.path.abspath(frame[1]) )
    return tmpfiles

def expandframelist(frames):
    '''Convert a dictionary with frames into a frame list where each frame
    gets its own entry in the form (tag, frame)
    '''
    framelist = list()
    for tag, f in frames:
        if isinstance(f, list) and not isinstance(f, fits.HDUList):
            framelist += [ (tag, frame) for frame in f ]
        elif f is not None:
            framelist.append((tag, f))
    return framelist




from __future__ import absolute_import
import os
import shutil
import tempfile
import threading
import collections
import warnings
import textwrap

from astropy.io import fits

from . import CPL_recipe
from .frames import FrameList, mkabspath, expandframelist
from .result import Result, RecipeCrash
from .param import ParameterList
from .logger import LogServer

class Recipe(object):
    '''Pluggable Data Reduction Module (PDRM) from a ESO pipeline. 

    Recipes are loaded from shared libraries that are provided with the
    pipeline library of the instrument. The module does not need to be
    linked to the same library version as the one used for the compilation
    of python-cpl. Currently, recipes compiled with CPL versions from 4.0
    are supported. The list of supported versions is stored as
    :attr:`cpl.cpl_versions`.

    The libraries are searched in the directories specified by the class
    attribute :attr:`Recipe.path` or its subdirectories. The search path is
    automatically set to the esorex path when :func:`cpl.esorex.init()`
    is called.
    '''

    path = [ '.' ]
    '''Search path for the recipes. It may be set to either a string, or to a
    list of strings. All shared libraries in the search path and their
    subdirectories are searched for CPL recipes. On default, the path is
    set to the current directory.

    The search path is automatically set to the esorex path when
    :func:`cpl.esorex.init()` is called.
    '''

    memory_mode = 0
    '''CPL memory management mode. The valid values are

    0
      Use the default system functions for memory handling

    1
      Exit if a memory-allocation fails, provide checking for memory leaks,
      limited reporting of memory allocation and limited protection on
      deallocation of invalid pointers.

    2
      Exit if a memory-allocation fails, provide checking for memory leaks,
      extended reporting of memory allocation and protection on deallocation
      of invalid pointers.

    .. note::

      This variable is only effective before the CPL library was
      initialized. Even :func:`cpl.Recipe.list()` initializes the library.
      Therefore it is highly recommended to set this as the first action after
      importing :mod:`cpl`.
    '''

    def __init__(self, name, filename = None, version = None, threaded = False):
        '''Try to load a recipe with the specified name in the directory
        specified by the class attribute :attr:`Recipe.path` or its 
        subdirectories.

        :param name: Name of the recipe. Required. Use
            :func:`cpl.Recipe.list()` to get a list of available recipes.
        :type name: :class:`str`

        :param filename:   Name of the shared library. Optional. If not set, 
            :attr:`Recipe.path` is searched for the library file. 
        :type filename: :class:`str`
        :param version: Version number. Optional. If not set, the newest 
            version is loaded.
        :type version: :class:`int` or :class:`str`
        :param threaded: Run the recipe in the background, returning 
            immediately after calling it. Default is :attr:`False`. This may 
            be also set as an attribute or specified as a parameter when 
            calling the recipe. 
        :type threaded: :class:`bool`

        '''
        os.putenv('CPL_MEMORY_MODE', str(Recipe.memory_mode));
        self._recipe = None
        self.__name__ = name
        '''Recipe name.'''

        if not filename:
            filename = Recipe.get_recipefilename(name, version)
            if not filename:
                raise IOError('Recipe %s not found at path %s'
                              % (repr(name), repr(Recipe.path)))
        self.__file__ = filename
        '''Shared library file name.'''

        self._recipe = CPL_recipe.recipe(filename, name)
        if version and version not in self.version:
            raise IOError('wrong version %s (requested %s) for %s in %s' %
                          (str(self.version), str(version), name, filename))
        if not self._recipe.cpl_is_supported():
            warnings.warn("Unsupported CPL version %s linked to %s" %
                          (self.cpl_version, filename))
        self._param = ParameterList(self)
        self._calib = FrameList(self)

        self.env = dict()
        '''Bla'''

        self._tags = None

        self.tag = self.tags[0] if self.tags else None
        '''Default tag when the recipe is called. This is set automatically
        only if the recipe provided the information about input
        tags. Otherwise this tag has to be set manually.
        '''

        self.output_dir = None

        self.temp_dir = '.'
        '''Base directory for temporary directories where the recipe is
        executed. The working dir is created as a subdir with a random file
        name. If set to :obj:`None`, the system temp dir is used.  Defaults to
        :literal:`'.'`.
        '''

        self.memory_dump = 0

        self.threaded = threaded

        self.mtrace = False

        self.__doc__ = self._doc()

    @property
    def __author__(self):
        '''Author name'''
        return self._recipe.author()[0]

    @property
    def __email__(self):
        '''Author email'''
        return self._recipe.author()[1]

    @property
    def description(self):
        '''Pair (synopsis, description) of two strings.'''
        return self._recipe.description()

    @property
    def version(self):
        '''Pair (versionnumber, versionstring) of an integer and a string. 
        The integer will be increased on development progress.'''
        return self._recipe.version()

    @property
    def __version__(self):
        return self._recipe.version()[1]

    @property
    def __copyright__(self):
        '''Copyright string of the recipe'''
        return self._recipe.copyright()

    @property
    def cpl_version(self):
        '''Version of the CPL library that is linked to the recipe,
        as a string'''
        return self._recipe.cpl_version()

    @property
    def cpl_description(self):
        '''Version numbers of CPL and its libraries that were linked to
        the recipe, as a string.'''
        return self._recipe.cpl_description()

    @property
    def tags(self):
        '''Possible tags for the raw input frames, or ':obj:`None` if this
        information is not provided by the recipe.'''
        frameconfig = self._recipe.frameConfig()
        return [ c[0][0] for c in frameconfig ] if frameconfig else self._tags

    @tags.setter
    def tags(self, t):
        if self._recipe.frameConfig():
            raise AttributeError('Tags are immutable')
        else:
            self._tags = t

    @property
    def tag(self):
        '''Default raw input frame tag. After creation, it is set to the first
        tag from the "tags" property and may be changed to any of these
        tags. If the recipe does not provide the tag information, it must be
        set manually, or the tag name has to be provided when calling the
        recipe.'''
        return self._tag

    @tag.setter
    def tag(self, t):
        if self.tags is None or t in self.tags:
            self._tag = t 
        else:
            raise KeyError("Tag %s not in %s" % (repr(t), str(self.tags)))

    @property
    def calib(self):
        '''This attribute contains the calibration frames
        for the recipe.  It is iterable and then returns all calibration frames:
        
        >>> for f in muse_scibasic.calib:
        ...     print f.tag, f.min, f.max, f.frames
        TRACE_TABLE 1 1 None
        WAVECAL_TABLE 1 1 None
        MASTER_BIAS 1 1 master_bias_0.fits
        MASTER_DARK None 1 None
        GEOMETRY_TABLE 1 1 None
        BADPIX_TABLE None None ['badpix_1.fits', 'badpix_2.fits']
        MASTER_FLAT None 1 None

        .. note::

           Only MUSE recipes are able to provide the full list of
           calibration frames and the minimal/maximal number of calibration
           frames. For other recipes, only frames that were set by the users are
           returned here. Their minimum and maximum value will be set to
           :obj:`None`.

        In order to assing a FITS file to a tag, the file name or the
        :class:`astropy.io.fits.HDUList` is assigned to the calibration
        attribute:

        >>> muse_scibasic.calib.MASTER_BIAS = 'MASTER_BIAS_0.fits'

        Using :class:`astropy.io.fits.HDUList` is useful when it needs to be
        patched before fed into the recipe.

        >>> master_bias = astropy.io.fits.open('MASTER_BIAS_0.fits')
        >>> master_bias[0].header['HIERARCH ESO DET CHIP1 OUT1 GAIN'] = 2.5
        >>> muse_scibasic.calib.MASTER_BIAS = master_bias

        Note that :class:`astropy.io.fits.HDUList` objects are stored in
        temporary files before the recipe is called which may produce some
        overhead. Also, the CPL then assigns the random temporary file names
        to the FITS keywords ``HIERARCH ESO PRO RECm RAWn NAME`` which should
        be corrected afterwards if needed.

        To assign more than one frame, put them into a list:

        >>> muse_scibasic.calib.BADPIX_TABLE = [ 'badpix1.fits', 'badpix2.fits' ]
     
        All calibration frames can be set in one step by assigning a
        :class:`dict` to the parameters. In this case, frame that are not in
        the map are set are removed from the list, and unknown frame tags are
        silently ignored. The key of the map is the tag name; the values are
        either a string, or a list of strings, containing the file name(s) or
        the :class:`astropy.io.fits.HDUList` objects.

        >>> muse_scibasic.calib = { 'MASTER_BIAS':'master_bias_0.fits', 
        ...                'BADPIX_TABLE':[ 'badpix_1.fits', 'badpix_2.fits' ] }

        In a recipe call, the calibration frame lists may be overwritten by
        specifying them in a :class:`dict`:

        >>> res = muse_scibasic( ..., calib = {'MASTER_BIAS':'master_bias_1.fits'})

        '''
        return self._calib

    @calib.setter
    def calib(self, source = None):
        if isinstance(source, str) or hasattr(source, 'read'):
            from .esorex import load_sof
            source = load_sof(source)
        self._calib = FrameList(self, source) 

    @calib.deleter
    def calib(self):
        self._calib = FrameList(self, None)

    @property
    def param(self):
        '''This attribute contains all recipe parameters. 
        It is iteratable and then returns all individual parameters:

        >>> for p in muse_scibasic.param:
        ...    print p.name, p.value, p.default
        ...
        nifu None 99
        cr None dcr
        xbox None 15
        ybox None 40
        passes None 2
        thres None 4.5
        sample None False
        dlambda None 1.2

        On interactive sessions, all parameter settings can be easily printed by
        printing the :attr:`param` attribute of the recipe:

        >>> print muse_scibasic.param
         [Parameter('nifu', default=99), Parameter('cr', default=dcr), 
          Parameter('xbox', default=15), Parameter('ybox', default=40), 
          Parameter('passes', default=2), Parameter('thres', default=4.5), 
          Parameter('sample', default=False), Parameter('dlambda', default=1.2)]

        To set the value of a recipe parameter, the value can be assigned to
        the according attribute:

        >>> muse_scibasic.param.nifu = 1

        The new value is checked against parameter type, and possible value
        limitations provided by the recipe. Hyphens in parameter names are
        converted to underscores. In a recipe call, the same parameter can be
        specified as :class:`dict`:

        >>> res = muse_scibasic( ..., param = {'nifu':1})

        To reset a value to its default, it is either deleted, or set to
        :obj:`None`. The following two lines:

        >>> muse_scibasic.param.nifu = None
        >>> del muse_scibasic.param.nifu

        will both reset the parameter to its default value. 

        All parameters can be set in one step by assigning a :class:`dict` to
        the parameters. In this case, all values that are not in the map are
        reset to default, and unknown parameter names are ignored. The keys of
        the map may contain contain the name or the fullname with context:

        >>> muse_scibasic.param = { 'nifu':1, 'xbox':11, 'resample':True }
        '''
        return self._param

    @param.setter
    def param(self, source = None):
        if isinstance(source, str) or hasattr(source, 'read'):
            from .esorex import load_rc
            source = load_rc(source)
        self._param = ParameterList(self, source)

    @param.deleter
    def param(self):
        self._param = ParameterList(self, None)

    @property
    def output(self):
        '''Return a dictionary of output frame tags.

        Keys are the tag names, values are the corresponding list of output
        tags. If the recipe does not provide this information, an exception is
        raised.
        '''
        return dict((c[0][0], c[2]) for c in self._recipe.frameConfig())

    def __call__(self, *data, **ndata):
        '''Call the recipes execution with a certain input frame.
        
        :param raw: Data input frames.
        :type raw: :class:`astropy.io.fits.HDUlist` or :class:`str` or a 
            :class:`list` of them, or :class:`dict`
        :param tag: Overwrite the :attr:`tag` attribute (optional).
        :type tag: :class:`str`
        :param threaded: overwrite the :attr:`threaded` attribute (optional).
        :type threaded: :class:`bool`
        :param loglevel: set the log level for python :mod:`logging` (optional).
        :type loglevel: :class:`int`
        :param logname: set the log name for the python
            :class:`logging.Logger` (optional, default is 'cpl.' + recipename).
        :type logname: :class:`str`
        :param output_dir: Set or overwrite the :attr:`output_dir` attribute.
            (optional)
        :type output_dir: :class:`str`
        :param param: overwrite the CPL parameters of the recipe specified
            as keys with their dictionary values (optional). 
        :type param: :class:`dict`
        :param calib: Overwrite the calibration frame lists for the tags 
            specified as keys with their dictionary values (optional).
        :type calib: :class:`dict`
        :param env: overwrite environment variables for the recipe call 
            (optional). 
        :type env: :class:`dict`
        :return: The object with the return frames as 
            :class:`astropy.io.fits.HDUList` objects
        :rtype: :class:`cpl.Result`
        :raise: :exc:`exceptions.ValueError` If the invocation parameters
                are incorrect.
        :raise: :exc:`exceptions.IOError` If the temporary directory could
                not be built, the recipe could not start or the files could not 
                be read/written.
        :raise: :exc:`cpl.CplError` If the recipe returns an error.
        :raise: :exc:`cpl.RecipeCrash` If the CPL recipe crashes with a
                SIGSEV or a SIGBUS

        .. note::

            If the recipe is executed in the background 
            (``threaded = True``) and an exception occurs, this exception is 
            raised whenever result fields are accessed.
        '''
        threaded = ndata.get('threaded', self.threaded)
        mtrace = ndata.get('mtrace', self.mtrace)
        loglevel = ndata.get('loglevel')
        logname = ndata.get('logname', 'cpl.%s' % self.__name__)
        output_dir = ndata.get('output_dir', self.output_dir)
        output_format = str if output_dir else fits.HDUList
        if output_dir is None:
            output_dir = tempfile.mkdtemp(dir = self.temp_dir, 
                                          prefix = self.__name__ + "-") 
        parlist = self.param._aslist(ndata.get('param'))
        raw_frames = self._get_raw_frames(*data, **ndata)
        if len(raw_frames) < 1:
            raise ValueError('No raw frames specified.')
        input_len = -1 if isinstance(raw_frames[0][1], fits.HDUList) else \
            len(raw_frames[0][1]) if isinstance(raw_frames[0][1], list) else -1
        calib_frames = self.calib._aslist(ndata.get('calib'))
        framelist = expandframelist(raw_frames + calib_frames)
        runenv = dict(self.env)
        runenv.update(ndata.get('env', dict()))
        logger = None
        delete = output_format == fits.HDUList
        try:
            if (not os.access(output_dir, os.F_OK)):
                os.makedirs(output_dir)
            mkabspath(framelist, output_dir)
            logger = LogServer(logname, loglevel)
        except:
            try:
                self._cleanup(output_dir, logger, delete)
            except:
                pass
            raise
        if not threaded:
            return self._exec(output_dir, parlist, framelist, runenv, 
                         input_len, logger, output_format, delete, mtrace)
        else:
            return  Threaded(
                self._exec, output_dir, parlist, framelist, runenv, 
                input_len, logger, output_format, delete, mtrace)

    def _exec(self, output_dir, parlist, framelist, runenv,
              input_len, logger, output_format, delete, mtrace):
        try:
            return Result(output_dir,
                          self._recipe.run(output_dir, parlist, framelist,
                                           list(runenv.items()), 
                                           logger.logfile, logger.level,
                                           self.memory_dump, mtrace),
                          input_len, logger, output_format)
        finally:
            self._cleanup(output_dir, logger, delete)

    def _get_raw_frames(self, *data, **ndata):
        '''Return the input frames.

        Returns a :class:`list` with (tag, the input frame(s)) pairs. Note
        that more than one input tag is not allowed here.
        '''
        data = list(data)
        if 'raw' in ndata:
            data.append(ndata['raw'])
        tag = ndata.get('tag', self.tag)
        m = { }
        for f in data:
            if isinstance(f, dict):
                m.update(f)
            elif tag is None:
                raise ValueError('No raw input tag')
            elif tag not in m:
                m[tag] = f
            elif isinstance(m[tag], list) \
                    and not isinstance(m[tag], fits.HDUList):
                m[tag].append(f)
            else:
                m[tag] = [ m[tag], f ]
        return list(m.items())

    def _cleanup(self, output_dir, logger, delete):
        try:
            bt = os.path.join(output_dir, 'recipe.backtrace-unprocessed')
            if os.path.exists(bt):
                with open(bt) as bt_file:
                    os.rename(bt, os.path.join(output_dir, 'recipe.backtrace'))
                    ex = RecipeCrash(bt_file)
                    ex.log(logger.logger)
                    raise ex

        finally:
            if delete:
                shutil.rmtree(output_dir)

    def _doc(self):
        s = '%s\n\n%s\n\n' % (textwrap.fill(self.description[0]),
                              textwrap.fill(self.description[1]))
        
        if len(self.param) > 0:
            r = 'Parameters:\n%s\n' % self._param.__doc__
        else:
            r = 'No parameters\n'
        if self._recipe.frameConfig() is not None:
            c = textwrap.fill(repr([f.tag for f in self.calib]),
                              initial_indent = 'Calibration frames: ',
                              subsequent_indent = ' ' * 21) + '\n\n'
        else:
            c = ''
        if self.tags is not None:
            t = 'Raw and product frames:\n'
            maxlen = max(len(f) for f in self.tags)
            for f in self.tags:
                t += textwrap.fill(repr(self.output[f]),
                                   initial_indent = ' %s --> ' % f.rjust(maxlen),
                                   subsequent_indent = ' ' * (maxlen + 7)) + '\n'
        else:
            t = ''
        return s + r + c + t + '\n\n'

    def __repr__(self):
        return 'Recipe(%s, version = %s)' % (repr(self.__name__), 
                                             repr(self.version[0]))

    @staticmethod
    def list():
        '''Return a list of recipes.
        
        Searches for all recipes in in the directory specified by the class
        attribute :attr:`Recipe.path` or its subdirectories. 
        '''
        os.putenv('CPL_MEMORY_MODE', str(Recipe.memory_mode));
        plugins = collections.defaultdict(list)
        for f in Recipe.get_libs():
            plugin_f = CPL_recipe.list(f)
            if plugin_f:
                for p in plugin_f:
                    plugins[p[0]].append(p[2])
        return list(plugins.items())

    @staticmethod
    def get_recipefilename(name, version = None):
        os.putenv('CPL_MEMORY_MODE', str(Recipe.memory_mode));
        filename = None
        rversion = -1
        for f in Recipe.get_libs():
            plugin_f = CPL_recipe.list(f)
            if plugin_f:
                for p in plugin_f:
                    if p[0] != name:
                        continue
                    if version in p[1:3]:
                        return f
                    if rversion < p[1]:
                        rversion = p[1]
                        filename = f
        return filename

    @staticmethod
    def get_libs():
        libs = [ ]
        path = Recipe.path.split(':') if isinstance(Recipe.path, str) else Recipe.path
        for p in path:
            for root, d, files in os.walk(p):
                libs += [ os.path.join(root, f) 
                           for f in files if f.endswith('.so') ]
        return libs

    @staticmethod
    def set_maxthreads(n):
        '''Set the maximal number of threads to be executed in parallel.

        .. note::

            This affects only threads that are started afterwards with
            the ``threaded = True`` flag.

        .. seealso:: :ref:`parallel`
        '''
        Threaded.set_maxthreads(n)

class Threaded(threading.Thread):
    '''Simple threading interface. 

    Creating this object will start the execution of func(*args, **nargs).
    It returns an object that has the same attribute as the function return
    value, if the function execution was completed.

    Accessing any of the attributes will cause a wait until the function
    execution is ready. Note that the attribute delegation will work only for
    attributes (not for methods), and it will not work for attributes defined
    by the threading.Thread interface.

    If the function returns an exception, this exception is thrown by any
    attempt to access an attribute.
    '''
    pool_sema = threading.BoundedSemaphore(65536)

    def __init__(self, func, *args, **nargs):
        threading.Thread.__init__(self)
        self._func = func
        self._args = args
        self._nargs = nargs
        self._res = None
        self._exception = None
        self.start()
                    
    def run(self):
        with Threaded.pool_sema:
            try:
                self._res = self._func(*self._args, **self._nargs)
            except Exception as exception:
                self._exception = exception

    @property
    def _result(self):
        self.join()
        if self._exception is None:
            return self._res
        else:
            raise self._exception

    def __getitem__(self, key):
        return self._result[key]

    def __contains__(self, key):
        return key in self._result.tags

    def __len__(self):
        return len(self._result.tags)

    def __iter__(self):
        return self._result.__iter__()

    def __getattr__(self, name):
        return self._result.__dict__[name]

    @staticmethod
    def set_maxthreads(n):
        with Threaded.pool_sema:
            Threaded.pool_sema = threading.BoundedSemaphore(n)




import hashlib

def datamd5(hdulist):
    '''Calculate the MD5SUM of all data regions of a HDUList.
    '''
    md5sum = hashlib.md5()
    for hdu in hdulist:
        if hdu.data is not None:
            md5sum.update(bytes(hdu.data.data))
            pad = b'\0' * (2880 - hdu.data.nbytes % 2880)
            md5sum.update(pad)
    return md5sum.hexdigest()

def verify_md5(hdulist):
    return hdulist[0].header.get('DATAMD5') == datamd5(hdulist)

def update_md5(hdulist):
    md5sum = datamd5(hdulist)
    hdulist[0].header['DATAMD5'] =  (md5sum, 'MD5 checksum')
    return md5sum



import logging
import os
import shutil
import tempfile
import unittest

import numpy
from astropy.io import fits
import cpl
cpl.Recipe.memory_mode = 0

recipe_name = 'rtest'
raw_tag = 'RRRECIPE_DOCATG_RAW'
rtest_esorexrc = '''# example RC file
iiinstrument.rtest.string_option = more
iiinstrument.rtest.bool_option = False
iiinstrument.rtest.float_option = 1.125
iiinstrument.rtest.int_option = -2
iiinstrument.rtest.enum_option = second
iiinstrument.rtest.range_option = 0.5
iiinstrument.rtest.dotted.opt = 1
'''

def create_recipe(name, builddir):
    var = {
        'CC':  os.getenv("CC", "gcc"),
        'CPPFLAGS': os.getenv("CPPFLAGS", ""),
        'CFLAGS': os.getenv("CFLAGS", ""),
        'LDFLAGS': os.getenv("LDFLAGS", ""),
        'LIBS': "-lcplcore -lcpldfs",
        'cname': os.path.join(os.path.dirname(__file__), name + ".c"),
        'oname': os.path.join(builddir, name + '.o'),
        'soname': os.path.join(builddir, name + '.so'),
    }
    os.system("{CC} {CPPFLAGS} {CFLAGS} -fPIC -c -o {oname} {cname}".format(**var))
    os.system("{CC} {LDFLAGS} -shared -o {soname} {oname} {LIBS}".format(**var))
    os.remove(var['oname'])

class CplTestCase(unittest.TestCase):
    def setUp(self):
        unittest.TestCase.setUp(self)
        self.temp_dir = tempfile.mkdtemp()
        create_recipe(recipe_name, self.temp_dir)
        cpl.Recipe.path = self.temp_dir

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        shutil.rmtree(self.temp_dir)

class RecipeTestCase(CplTestCase):
    def setUp(self):
        CplTestCase.setUp(self)
        self.recipe = cpl.Recipe(recipe_name)
        self.recipe.temp_dir = self.temp_dir
        self.recipe.tag = raw_tag
        self.image_size = (16, 16)
        self.raw_frame = fits.HDUList([
                fits.PrimaryHDU(numpy.random.randint(0, 65001,
                                                     self.image_size))])
        self.raw_frame[0].header['HIERARCH ESO DET DIT'] = 0.0
        self.raw_frame[0].header['HIERARCH ESO PRO CATG'] = raw_tag

class RecipeStatic(CplTestCase):
    def test_list(self):
        '''List available recipes'''
        l = cpl.Recipe.list()
        self.assertTrue(isinstance(l, list))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0], (recipe_name, ['0.0.1']))

    def test_create_recipe(self):
        '''Create a recipe specified by its name'''
        recipe = cpl.Recipe(recipe_name)
        self.assertTrue(isinstance(recipe, cpl.Recipe))

    def test_create_recipe_version(self):
        '''Create a recipe specified by its name and version'''
        recipe = cpl.Recipe(recipe_name, version = '0.0.1')
        self.assertTrue(isinstance(recipe, cpl.Recipe))        

    def test_create_recipe_wrong_name(self):
        '''Create a recipe specified by a wrong name'''
        self.assertRaises(IOError, cpl.Recipe, 'wrongname')

    def test_create_recipe_wrong_version(self):
        '''Create a recipe specified by a wrong version'''
        self.assertRaises(IOError, cpl.Recipe, recipe_name, version='0.0.10')

    def test_create_recipe_filename(self):
        '''Create a recipe specified by a the name and the filename'''
        recipe = cpl.Recipe(recipe_name,
                            filename = os.path.join(self.temp_dir, 'rtest.so'))
        self.assertTrue(isinstance(recipe, cpl.Recipe))

    def test_create_recipe_wrong_filename(self):
        '''Create a recipe specified by a wrong filename'''
        self.assertRaises(IOError, cpl.Recipe, recipe_name, 
                          filename = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'iiinstrumentp', 'recipes', '.libs', 'rtest.o'))

class RecipeCommon(RecipeTestCase):
    def test_name(self):
        '''Recipe name'''
        self.assertEqual(self.recipe.__name__, recipe_name)

    def test_author(self):
        '''Author attribute'''
        self.assertEqual(self.recipe.__author__, 'Ole Streicher')

    def test_email(self):
        '''Author attribute'''
        self.assertEqual(self.recipe.__email__, 'python-cpl@liska.ath.cx')

    def test_description(self):
        '''Synopsis and description'''
        self.assertTrue(isinstance(self.recipe.description[0], str))
        self.assertTrue(len(self.recipe.description[0]) > 0)
        self.assertTrue(isinstance(self.recipe.description[1], str))
        self.assertTrue(len(self.recipe.description[1]) > 0)

    def test_copyright(self):
        '''Copyright'''
        self.assertTrue(isinstance(self.recipe.__copyright__, str))
        self.assertTrue(len(self.recipe.__copyright__) > 0)

class RecipeParams(RecipeTestCase):
    def test_string_parameter(self):
        '''String parameter'''
        self.assertTrue(isinstance(self.recipe.param.stropt, cpl.Parameter))
        self.assertEqual(self.recipe.param.stropt.name, 'stropt')
        self.assertEqual(self.recipe.param.stropt.context, 
                         'iiinstrument.rtest')
        self.assertEqual(self.recipe.param.stropt.default, None)
        self.assertEqual(self.recipe.param.stropt.value, None)
        self.assertEqual(self.recipe.param.stropt.range, None)
        self.assertEqual(self.recipe.param.stropt.sequence, None)
        self.recipe.param.stropt = 'more'
        self.assertEqual(self.recipe.param.stropt.value, 'more')
        del self.recipe.param.stropt 
        self.assertEqual(self.recipe.param.stropt.value, None)

    def test_boolean_parameter(self):
        '''Boolean parameter'''
        self.assertTrue(isinstance(self.recipe.param.boolopt, cpl.Parameter))
        self.assertEqual(self.recipe.param.boolopt.name, 'boolopt')
        self.assertEqual(self.recipe.param.boolopt.default, True)
        self.assertEqual(self.recipe.param.boolopt.value, None)
        self.recipe.param.boolopt = False
        self.assertEqual(self.recipe.param.boolopt.value, False)
        del self.recipe.param.boolopt 
        self.assertEqual(self.recipe.param.boolopt.value, None)
        
    def test_float_parameter(self):
        '''Float parameter'''
        self.assertTrue(isinstance(self.recipe.param.floatopt, cpl.Parameter))
        self.assertEqual(self.recipe.param.floatopt.name, 'floatopt')
        self.assertEqual(self.recipe.param.floatopt.default, 0.1)
        self.assertEqual(self.recipe.param.floatopt.value, None)
        self.recipe.param.floatopt = 1.1
        self.assertEqual(self.recipe.param.floatopt.value, 1.1)
        del self.recipe.param.floatopt 
        self.assertEqual(self.recipe.param.floatopt.value, None)

    def test_int_parameter(self):
        '''Integer parameter'''
        self.assertTrue(isinstance(self.recipe.param.intopt, cpl.Parameter))
        self.assertEqual(self.recipe.param.intopt.name, 'intopt')
        self.assertEqual(self.recipe.param.intopt.default, 2)
        self.assertEqual(self.recipe.param.intopt.value, None)
        self.recipe.param.intopt = -1
        self.assertEqual(self.recipe.param.intopt.value, -1)
        del self.recipe.param.intopt 
        self.assertEqual(self.recipe.param.intopt.value, None)

    def test_enum_parameter(self):
        '''Enumeration (string) parameter'''
        self.assertTrue(isinstance(self.recipe.param.enumopt, cpl.Parameter))
        self.assertEqual(self.recipe.param.enumopt.name, 'enumopt')
        self.assertEqual(self.recipe.param.enumopt.default, 'first')
        self.assertEqual(self.recipe.param.enumopt.value, None)
        self.recipe.param.enumopt = 'second'
        self.assertEqual(self.recipe.param.enumopt.value, 'second')
        del self.recipe.param.enumopt 
        self.assertEqual(self.recipe.param.enumopt.value, None)
        def setenumoptinvalid():
            self.recipe.param.enumopt = 'invalid'
        self.assertRaises(ValueError, setenumoptinvalid)

    def test_range_parameter(self):
        '''Range (float) parameter'''
        self.assertTrue(isinstance(self.recipe.param.rangeopt, cpl.Parameter))
        self.assertEqual(self.recipe.param.rangeopt.name, 'rangeopt')
        self.assertEqual(self.recipe.param.rangeopt.default, 0.1)
        self.assertEqual(self.recipe.param.rangeopt.value, None)
        self.recipe.param.rangeopt = 0.4
        self.assertEqual(self.recipe.param.rangeopt.value, 0.4)
        del self.recipe.param.rangeopt 
        self.assertEqual(self.recipe.param.rangeopt.value, None)
        def setrangeoptinvalid():
            self.recipe.param.rangeopt = 1.5
        self.assertRaises(ValueError, setrangeoptinvalid)

    def test_as_dict(self):
        '''Use the parameter list as a dictionary'''
        self.assertEqual(self.recipe.param.boolopt, 
                         self.recipe.param['boolopt'])
        self.assertEqual(self.recipe.param.boolopt, 
                         self.recipe.param['iiinstrument.rtest.bool_option'])

    def test_dotted_par(self):
        '''Use a parameter that has a dot in its alias'''
        self.assertEqual(self.recipe.param.dot.opt,
                         self.recipe.param.dot['opt'])
        self.assertEqual(self.recipe.param.dot.opt,
                         self.recipe.param['dot.opt'])
        self.assertEqual(self.recipe.param.dot.opt,
                         self.recipe.param['iiinstrument.rtest.dotted.opt'])

    def test_iterate(self):
        '''Iteration over all parameters'''
        for p in self.recipe.param:
            self.assertTrue(isinstance(p, cpl.Parameter))
        pars = [p.name for p in self.recipe.param]
        self.assertEqual(len(pars), len(self.recipe.param))
        self.assertTrue('stropt' in pars)
        self.assertTrue('boolopt' in pars)

    def test_set_dict(self):
        '''Assign a dictionary to the parameter list'''
        self.recipe.param = { 'stropt':'dmore', 'boolopt':True }
        self.assertEqual(self.recipe.param.boolopt.value, True)
        self.assertEqual(self.recipe.param.stropt.value, 'dmore')

        # Check that we can assign a dictionary with the short names and string
        self.recipe.param = { 'stropt':'dmore', 'boolopt':'False' }
        self.assertEqual(self.recipe.param.boolopt.value, False)

        # Check that we can assign a dictionary with the long names
        self.recipe.param = { 'iiinstrument.rtest.string_option':'dless', 
                      'iiinstrument.rtest.float_option':1.5, 
                      'iiinstrument.rtest.bool_option':True }
        self.assertEqual(self.recipe.param.stropt.value, 'dless')
        self.assertEqual(self.recipe.param.floatopt.value, 1.5)
        self.assertEqual(self.recipe.param.boolopt.value, True)

    def test_param_load_from_esorexrc(self):
        '''Load parameters from an Esorex .rc file'''
        self.recipe.param = rtest_esorexrc
        self.assertEqual(self.recipe.param.stropt.value, 'more')
        self.assertEqual(self.recipe.param.boolopt.value, False)
        self.assertEqual(self.recipe.param.floatopt.value, 1.125)
        self.assertEqual(self.recipe.param.intopt.value, -2)
        self.assertEqual(self.recipe.param.enumopt.value, 'second')
        self.assertEqual(self.recipe.param.rangeopt.value, 0.5)
        self.assertEqual(self.recipe.param.dot.opt.value, 1)

    def test_delete(self):
        '''Delete all parameter values to reset to default'''
        self.recipe.param.boolopt.value = True
        self.recipe.param.stropt.value = 'something'
        del self.recipe.param
        self.assertEqual(self.recipe.param.stropt.value, None)
        self.assertEqual(self.recipe.param.boolopt.value, None)

    def test_dir(self):
        '''[TAB] completition. 
        This requires to have   the __dir__() method working.
        '''
        self.assertEqual(set(self.recipe.param.__dir__()),
                         set(p.name if '.' not in p.name
                             else p.name.split('.', 1)[0]
                             for p in self.recipe.param
                             ))

    def test_str(self):
        '''Parameter string representation.
        
        Since the string depends on the order of parameter, we check
        that it evaluates to the correct dict.
        '''
        self.assertEqual(eval(str(self.recipe.param)), {
            'stropt': None,
            'boolopt': True,
            'floatopt': 0.1,
            'intopt': 2,
            'enumopt': 'first',
            'rangeopt': 0.1,
            'crashing': 'no',
            'memleak': False,
            'sleep': 0.1,
            'disabled': -0.1,
            'dot.opt': 0})
        
    def test_repr(self):
        '''Canonical parameter string representation.
        
        Since the string depends on the order of parameter, we check
        that it evaluates to the correct dict.
        '''
        self.assertEqual(eval(repr(self.recipe.param)), {
            'stropt': None,
            'boolopt': True,
            'floatopt': 0.1,
            'intopt': 2,
            'enumopt': 'first',
            'rangeopt': 0.1,
            'crashing': 'no',
            'memleak': False,
            'sleep': 0.1,
            'disabled': -0.1,
            'dot.opt': 0})

    def test_eq(self):
        '''Trivial equality test'''
        self.assertTrue(self.recipe.param == self.recipe.param)

        
class RecipeCalib(RecipeTestCase):
    def test_set(self):
        '''Set a calibration frame'''
        self.recipe.calib.FLAT = 'flat.fits'
        self.assertEqual(self.recipe.calib.FLAT.frames, 'flat.fits')
        
    def test_set_dict(self):
        '''Assign a dictionary to the calibration frame list'''
        self.recipe.calib = { 'FLAT':'flat2.fits' }
        self.assertEqual(self.recipe.calib.FLAT.frames, 'flat2.fits')

    def test_param_load_from_esorexsof(self):
        '''Load calibration from an Esorex SOF file'''
        self.recipe.calib = 'flat.fits   FLAT\n'
        self.assertEqual(self.recipe.calib.FLAT.frames, 'flat.fits')

    def test_del(self):
        '''Delete a calibration frame set'''
        self.recipe.calib.FLAT = 'flat.fits'
        del self.recipe.calib.FLAT
        f = self.recipe.calib.FLAT.frames
        self.assertEqual(f, None)

    def test_del_all(self):
        '''Delete all calibration frame sets'''
        self.recipe.calib.FLAT = 'flat.fits'
        del self.recipe.calib
        try:
            f = self.recipe.calib.FLAT.frames
        except:
            f = None
        self.assertEqual(f, None)

    def test_dir(self):
        '''[TAB] completition. 
        This requires to have   the __dir__() method working.
        '''
        self.assertEqual(set(self.recipe.calib.__dir__()),
                         set(f.tag for f in self.recipe.calib))

    def test_str(self):
        '''Calibration frames string representation'''
        self.assertEqual(str(self.recipe.calib),
                             "{}")
        
    def test_repr(self):
        '''Canonical calibration frames string representation'''
        self.assertEqual(repr(self.recipe.calib),
                             "{}")

    def test_eq(self):
        '''Trivial equality test'''
        self.assertTrue(self.recipe.calib == self.recipe.calib)

class RecipeExec(RecipeTestCase):
    def setUp(self):
        RecipeTestCase.setUp(self)
        self.flat_frame = fits.HDUList([
                fits.PrimaryHDU(numpy.random.randint(0, 65001,
                                                     self.image_size))])

    def test_frames_keyword_dict(self):
        '''Raw and calibration frames specified as keyword dict'''
        self.recipe.tag = None
        res = self.recipe(raw = {'RRRECIPE_DOCATG_RAW': self.raw_frame },
                          calib = { 'FLAT':self.flat_frame })
        self.assertTrue(isinstance(res, cpl.Result))
        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
        self.assertTrue(abs(self.raw_frame[0].data 
                            - res.THE_PRO_CATG_VALUE[0].data).max() == 0)
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass

    def test_frames_keyword_calib(self):
        '''Raw frame specified as keyword, calibration frame set in recipe'''
        self.recipe.tag = None
        self.recipe.calib.FLAT = self.flat_frame
        res = self.recipe({'RRRECIPE_DOCATG_RAW':self.raw_frame})
        self.assertTrue(isinstance(res, cpl.Result))
        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass

    def test_frames_tag_keyword(self):
        '''The 'tag' parameter'''
        self.recipe.tag = None
        self.recipe.calib.FLAT = self.flat_frame
        res = self.recipe(self.raw_frame, tag = raw_tag)
        self.assertTrue(isinstance(res, cpl.Result))
        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass

    def test_frames_tag_attribute(self):
        '''The 'tag' attribute'''
        self.recipe.tag = raw_tag
        res = self.recipe(self.raw_frame)
        self.assertTrue(isinstance(res, cpl.Result))
        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass

    def test_frames_one_element_input_list(self):
        '''Use 1-element list as input'''
        # --> we want a list back'''
        res = self.recipe([self.raw_frame])
        self.assertTrue(isinstance(res, cpl.Result))
        self.assertFalse(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, list))
        try:
            res.THE_PRO_CATG_VALUE[0].close()
        except:
            pass

    def test_frames_many_element_input_list(self):
        '''Use multiple files as input'''
        # --> since we only get back one image, it is
        # assumed to be a 'master', and we get back a plain frame'''
        res = self.recipe([self.raw_frame, self.raw_frame])
        self.assertTrue(isinstance(res, cpl.Result))
        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass

    def test_output_dir_attribute(self):
        '''Write an output dir specified as attribute'''
        output_dir = os.path.join(self.temp_dir, 'out')
        self.recipe.output_dir = output_dir
        res = self.recipe(self.raw_frame)
        self.assertTrue(isinstance(res, cpl.Result))
        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, str))
        self.assertEqual(os.path.basename(res.THE_PRO_CATG_VALUE), 
                         'rtest.fits')
        self.assertTrue(os.path.isdir(output_dir))
        self.assertTrue(os.path.isfile(res.THE_PRO_CATG_VALUE))
        hdu = fits.open(res.THE_PRO_CATG_VALUE)
        self.assertTrue(isinstance(hdu, fits.HDUList))
        try:
            hdu.close()
        except:
            pass

    def test_output_dir_keyword(self):
        '''Write an output dir specified as call keyword arg'''
        output_dir = os.path.join(self.temp_dir, 'out')
        res = self.recipe(self.raw_frame, output_dir = output_dir)
        self.recipe.output_dir = output_dir
        res = self.recipe(self.raw_frame)
        self.assertTrue(os.path.isdir(output_dir))
        self.assertTrue(isinstance(res, cpl.Result))
        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, str))
        self.assertEqual(os.path.basename(res.THE_PRO_CATG_VALUE), 
                         'rtest.fits')
        self.assertTrue(os.path.isfile(res.THE_PRO_CATG_VALUE))
        hdu = fits.open(res.THE_PRO_CATG_VALUE)
        self.assertTrue(isinstance(hdu, fits.HDUList))
        try:
            hdu.close()
        except:
            pass

    def test_param_default(self):
        '''Test default parameter settings'''
        res = self.recipe(self.raw_frame).THE_PRO_CATG_VALUE
        self.assertEqual(res[0].header['HIERARCH ESO QC STROPT'].strip(),
                         self.recipe.param.stropt.default or '')
        self.assertEqual(res[0].header['HIERARCH ESO QC BOOLOPT'],
                         self.recipe.param.boolopt.default)
        self.assertEqual(res[0].header['HIERARCH ESO QC INTOPT'],
                         self.recipe.param.intopt.default)
        self.assertEqual(res[0].header['HIERARCH ESO QC FLOATOPT'],
                         self.recipe.param.floatopt.default)
        self.assertEqual(res[0].header['HIERARCH ESO QC ENUMOPT'],
                         self.recipe.param.enumopt.default)
        self.assertEqual(res[0].header['HIERARCH ESO QC RANGEOPT'],
                         self.recipe.param.rangeopt.default)
        try:
            res.close()
        except:
            pass

    def test_param_keyword_dict(self):
        '''Parameter handling via keyword dict'''
        res = self.recipe(self.raw_frame, 
                          param = { 'stropt':'more' }).THE_PRO_CATG_VALUE
        self.assertEqual(res[0].header['HIERARCH ESO QC STROPT'], 'more')
        try:
            res.close()
        except:
            pass

    def test_param_keyword_dict_wrong(self):
        '''Parameter handling via keyword dict'''
        self.assertRaises(KeyError, self.recipe,
                          self.raw_frame, param = { 'wrong':True })

    def test_param_setting(self):
        '''Parameter handling via recipe setting'''
        self.recipe.param.stropt = 'more'
        with self.recipe(self.raw_frame).THE_PRO_CATG_VALUE as res:
            self.assertEqual(res[0].header['HIERARCH ESO QC STROPT'], 'more')

    def test_param_delete(self):
        '''Delete a parameter in a second run after setting it'''
        self.recipe.param.intopt = 123
        with self.recipe(self.raw_frame).THE_PRO_CATG_VALUE as res:
            pass
        del self.recipe.param.intopt
        with self.recipe(self.raw_frame).THE_PRO_CATG_VALUE as res:
            self.assertEqual(res[0].header['HIERARCH ESO QC INTOPT'], 2)

    def test_param_overwrite(self):
        '''Overwrite the recipe setting param via via keyword arg'''
        self.recipe.param.stropt = 'more'
        res = self.recipe(self.raw_frame, param = {'stropt':'less'}).THE_PRO_CATG_VALUE
        self.assertEqual(res[0].header['HIERARCH ESO QC STROPT'], 'less')

    def test_param_types(self):
        '''Parameter types'''
        self.recipe.param.stropt = 'more'
        self.recipe.param.boolopt = False
        self.recipe.param.intopt = 123
        self.recipe.param.floatopt = -0.25
        self.recipe.param.enumopt = 'third'
        self.recipe.param.rangeopt = 0.125
        with self.recipe(self.raw_frame).THE_PRO_CATG_VALUE as res:
            self.assertEqual(res[0].header['HIERARCH ESO QC STROPT'], 'more')
            self.assertEqual(res[0].header['HIERARCH ESO QC BOOLOPT'], False)
            self.assertEqual(res[0].header['HIERARCH ESO QC INTOPT'], 123)
            self.assertEqual(res[0].header['HIERARCH ESO QC FLOATOPT'], -0.25)
            self.assertEqual(res[0].header['HIERARCH ESO QC ENUMOPT'], 'third')
            self.assertEqual(res[0].header['HIERARCH ESO QC RANGEOPT'], 0.125)
        
    def test_disabled(self):
        '''Parameter with CLI disabled'''
        self.assertFalse(self.recipe.param.disabled.enabled[0])
        self.assertTrue(self.recipe.param.intopt.enabled[0])
#        self.recipe.param.disabled = 0.2
#        res = self.recipe(self.raw_frame)
#        self.assertEqual(res[0].header['HIERARCH ESO QC DISABLED'], 0.2)
#        try:
#            res.close()
#        except:
#            pass

    def test_environment_setting(self):
        '''Additional environment parameter via recipe setting'''
        self.recipe.env['TESTENV'] = 'unkk'
        with self.recipe(self.raw_frame).THE_PRO_CATG_VALUE as res:
            self.assertEqual(res[0].header['HIERARCH ESO QC TESTENV'], 'unkk')

    def test_environment_keyword(self):
        '''Additional environment parameter via recipe call keyword'''
        with self.recipe(self.raw_frame, 
                         env = {'TESTENV':'kknu'}).THE_PRO_CATG_VALUE as res:
            self.assertEqual(res[0].header['HIERARCH ESO QC TESTENV'], 'kknu')

    def test_error(self):
        '''Error handling'''
        self.recipe.tag = 'some_unknown_tag'
        self.assertRaises(cpl.CplError, self.recipe, self.raw_frame)

    def test_parallel(self):
        '''Parallel execution'''
        results = list()
        for i in range(20):
            # mark each frame so that we can see their order
            self.raw_frame[0].header['HIERARCH ESO RAW1 NR'] = i
            results.append(self.recipe(self.raw_frame, param = {'intopt':i},
                                       env = {'TESTENV':('knu%02i' % i)},
                                       threaded = True))
        for i, res in enumerate(results):
            # check if we got the correct type
            self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
            # check if we have the correct parameter
            self.assertEqual(res.THE_PRO_CATG_VALUE[0].header[
                    'HIERARCH ESO QC INTOPT'], i)
            # check if we have the correct environment
            self.assertEqual(res.THE_PRO_CATG_VALUE[0].header[
                    'HIERARCH ESO QC TESTENV'], ('knu%02i' % i))
            # check if we have the correct input frame
            self.assertEqual(res.THE_PRO_CATG_VALUE[0].header[
                    'HIERARCH ESO RAW1 NR'], i)
            # check that the data were moved correctly
            self.assertTrue(abs(self.raw_frame[0].data 
                                - res.THE_PRO_CATG_VALUE[0].data).max() < 1e-6)
            try:
                res.THE_PRO_CATG_VALUE.close()
            except:
                pass

    def test_error_parallel(self):
        '''Error handling in parallel execution'''
        self.recipe.tag = 'some_unknown_tag'
        res = self.recipe(self.raw_frame, threaded = True)
        def get(x):
            return x.THE_PRO_CATG_VALUE
        self.assertRaises(cpl.CplError, get, res)

    def test_md5sum_result(self):
        '''MD5sum of the result file'''
        self.recipe.tag = raw_tag
        res = self.recipe(self.raw_frame)
        key = 'DATAMD5'
        md5sum = res.THE_PRO_CATG_VALUE[0].header[key]
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass
        self.assertNotEqual(md5sum, 'Not computed')
        self.assertEqual(len(md5sum), 
                         len('9d123996fa9a7bda315d07e063043454'))

    def test_md5sum_calib(self):
        '''Created MD5sum for a HDUList calib file'''
        self.recipe.tag = raw_tag
        self.recipe.calib.FLAT = self.flat_frame
        res = self.recipe(self.raw_frame)
        key = 'HIERARCH ESO PRO REC1 CAL1 DATAMD5'
        md5sum = res.THE_PRO_CATG_VALUE[0].header[key]
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass
        self.assertNotEqual(md5sum, 'Not computed')
        self.assertEqual(len(md5sum), 
                         len('9d123996fa9a7bda315d07e063043454'))

class RecipeCrashing(RecipeTestCase):
    def _test_corrupted(self):
        '''Handling of recipe crashes because of corrupted memory'''
        self.recipe.param.crashing = 'free'
        self.assertRaises(cpl.RecipeCrash, self.recipe, self.raw_frame)

    def _test_segfault(self):
        '''Handling of recipe crashes because of segmentation fault'''
        self.recipe.param.crashing = 'segfault'
        self.assertRaises(cpl.RecipeCrash, self.recipe, self.raw_frame)

    def _test_cleanup_after_crash(self):
        '''Test that a second run after a crash will succeed'''
        output_dir = os.path.join(self.temp_dir, 'out')
        self.recipe.output_dir = output_dir
        self.recipe.param.crashing = 'segfault'
        self.assertRaises(cpl.RecipeCrash, self.recipe, self.raw_frame)
        del self.recipe.param.crashing
        self.recipe(self.raw_frame)

class RecipeRes(RecipeTestCase):
    def setUp(self):
        RecipeTestCase.setUp(self)
        self.res = self.recipe(self.raw_frame)

    def tearDown(self):
        RecipeTestCase.tearDown(self)
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass

    def test_attribute(self):
        '''The result as an attribute'''
        self.assertTrue(isinstance(self.res.THE_PRO_CATG_VALUE, 
                                   fits.HDUList))

    def test_dict(self):
        '''The result as an attribute'''
        self.assertTrue(isinstance(self.res['THE_PRO_CATG_VALUE'], 
                                   fits.HDUList))
    def test_in(self):
        '''Check whether a tag is part of the result'''
        self.assertTrue('THE_PRO_CATG_VALUE' in self.res)
        self.assertFalse('Anothervalue' in self.res)

    def test_keyerror(self):
        '''Accessing an inexisting value'''
        self.assertRaises(KeyError, lambda val: self.res[val], 'Anothervalue')

    def test_len(self):
        '''Length of the result'''
        self.assertEqual(len(self.res), 1)

    def test_iter(self):
        '''Iterate over the result'''
        for tag, hdu in self.res:
            self.assertEqual(tag, 'THE_PRO_CATG_VALUE')
            self.assertTrue(isinstance(hdu, fits.HDUList))

class RecipeEsorex(CplTestCase):
    def tearDown(self):
        CplTestCase.tearDown(self)
        cpl.esorex.msg.level = cpl.esorex.msg.OFF
        cpl.esorex.log.level = cpl.esorex.msg.OFF

    def test_read_sof(self):
        '''Read a SOF file'''
        soffile = 'geometry_table1.fits GEOMETRY_TABLE\n' \
            'geometry_table2.fits GEOMETRY_TABLE\n' \
            'MASTER_BIAS-01.fits MASTER_BIAS\n' \
            'MASTER_FLAT-01.fits MASTER_FLAT\n' \
            '#sky_fullmoon_1.fits          SKY\n' \
            'sky_fullmoon_2.fits          SKY\n'
        self.assertEqual(cpl.esorex.load_sof(soffile),
                         { 'GEOMETRY_TABLE': ['geometry_table1.fits',
                                              'geometry_table2.fits' ],
                           'MASTER_BIAS': 'MASTER_BIAS-01.fits',
                           'MASTER_FLAT': 'MASTER_FLAT-01.fits',
                           'SKY': 'sky_fullmoon_2.fits' })
                           
    def test_read_rc(self):
        '''Read an EsoRec .rc file'''
        self.assertEqual(cpl.esorex.load_rc(rtest_esorexrc), 
                         { 'iiinstrument.rtest.string_option': 'more',
                           'iiinstrument.rtest.bool_option': 'False',
                           'iiinstrument.rtest.float_option': '1.125',
                           'iiinstrument.rtest.int_option': '-2',
                           'iiinstrument.rtest.enum_option': 'second',
                           'iiinstrument.rtest.range_option': '0.5',
                           'iiinstrument.rtest.dotted.opt': '1',
                       })
        
    def test_esorex_init(self):
        '''Init CPL from an esorex.rc file'''
        rcfile = '''esorex.caller.recipe-dir=/some/dir
        esorex.caller.msg-level=debug
        esorex.caller.log-level=info
        esorex.caller.log-dir=%s
        esorex.caller.log-file=some.log''' % self.temp_dir
        cpl.esorex.init(rcfile)
        self.assertEqual(cpl.esorex.msg.level, cpl.esorex.msg.DEBUG)
        self.assertEqual(cpl.esorex.log.level, cpl.esorex.msg.INFO)
        self.assertEqual(cpl.esorex.log.dir, self.temp_dir)
        self.assertEqual(cpl.esorex.log.filename, 'some.log')
        self.assertEqual(cpl.Recipe.path, ['/some/dir'])

    def test_esorex_log(self):
        '''Write a logfile controlled by the convienence logger'''
        dirname = os.path.join(self.temp_dir, 'log')
        filename = 'python-cpl.log'
        log_msg = 'Esorex convienence log'
        os.mkdir(dirname)
        cpl.esorex.log.dir = dirname
        cpl.esorex.log.filename = filename
        cpl.esorex.log.level = cpl.esorex.log.INFO
        filename = os.path.join(dirname, filename)
        logging.getLogger('cpl').info(log_msg)
        self.assertTrue(os.path.exists(filename))
        logfile = open(filename)
        log_content = logfile.read()
        logfile.close()
        self.assertTrue(log_msg in log_content)
        self.assertTrue('INFO' in log_content)

    def test_esorex_log_off(self):
        '''Switch the logfile off after writing something'''
        dirname = os.path.join(self.temp_dir, 'log')
        filename = 'python-cpl_off.log'
        log_msg = 'Esorex convienence log'
        os.mkdir(dirname)
        cpl.esorex.log.dir = dirname
        cpl.esorex.log.filename = 'python-cpl_debug.log'
        cpl.esorex.log.level = 'debug'
        logging.getLogger('cpl').debug(log_msg)
        cpl.esorex.log.filename = filename
        cpl.esorex.log.level = 'off'
        logging.getLogger('cpl').debug(log_msg)
        filename = os.path.join(dirname, filename)
        logfile = open(filename)
        log_content = logfile.read()
        logfile.close()
        self.assertEqual(len(log_content), 0)

class RecipeLog(RecipeTestCase):
    def setUp(self):
        RecipeTestCase.setUp(self)
        self.handler = RecipeLog.THandler()
        logging.getLogger('cpl.rtest').addHandler(self.handler)
        self.other_handler = RecipeLog.THandler()
        logging.getLogger('othername').addHandler(self.other_handler)

    def tearDown(self):
        RecipeTestCase.tearDown(self)
        logging.getLogger('cpl.rtest').removeHandler(self.handler)
        logging.getLogger('othername').removeHandler(self.other_handler)

    class THandler(logging.Handler):
        def __init__(self):
            logging.Handler.__init__(self)
            self.logs = list()

        def emit(self, record):
            self.logs.append(record)

        def clear(self):
            self.logs = list()

    def test_logging_DEBUG(self):
        '''Injection of CPL messages into the python logging system'''
        self.handler.clear()
        logging.getLogger().setLevel(logging.DEBUG)
        res = self.recipe(self.raw_frame)
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass

        # check that the logs are not empty
        self.assertNotEqual(len(self.handler.logs), 0)
        funcnames = set()
        lognames = set()
        for r in self.handler.logs:
            # Check that we saved the right class
            self.assertTrue(isinstance(r, logging.LogRecord))
            # Check that a message was provided
            self.assertNotEqual(r.msg, None)
            # Check that a function name was provided
            self.assertNotEqual(r.funcName, None)
            funcnames.add(r.funcName)
            lognames.add(r.name)
        # Check that we had at least one expected entry
        self.assertTrue('cpl_dfs_product_save' in funcnames)
        self.assertTrue('cpl.rtest.cpl_dfs_product_save' in lognames)
        
    def test_logging_INFO(self):
        '''Filtering INFO messages'''
        self.handler.clear()
        logging.getLogger('cpl.rtest').setLevel(logging.INFO)
        res = self.recipe(self.raw_frame)
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass

        # check that the logs are not empty
        self.assertNotEqual(len(self.handler.logs), 0)

    def test_logging_WARN(self):
        '''Filtering WARN messages'''
        self.handler.clear()
        logging.getLogger('cpl.rtest').setLevel(logging.WARN)
        res = self.recipe(self.raw_frame)
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass

        # check that the logs are not empty
        self.assertNotEqual(len(self.handler.logs), 0)

    def test_logging_ERROR(self):
        '''Filtering of error messages'''
        # There is no error msg written by the recipe, so it should be empty.
        self.handler.clear()
        logging.getLogger('cpl.rtest').setLevel(logging.ERROR)
        res = self.recipe(self.raw_frame)
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass
        self.assertEqual(len(self.handler.logs), 0)

    def test_logging_common(self):
        '''Log name specification on recipe call'''
        self.handler.clear()
        self.other_handler.clear()
        res = self.recipe(self.raw_frame, logname = 'othername')
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass
        self.assertNotEqual(len(self.other_handler.logs), 0)

    def test_logging_multiline(self):
        '''Multiple lines in messages'''
        self.handler.clear()
        logging.getLogger('cpl.rtest').setLevel(logging.INFO)
        res = self.recipe(self.raw_frame)
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass
        # check that the multi line log sequence appears
        multiline = 0
        tag = 'multiline#'
        for l in self.handler.logs:
            if tag not in l.msg:
                continue
            i = int(l.msg[l.msg.index(tag)+len(tag):].split()[0])
            self.assertEqual(multiline + 1, i)
            multiline = i
        self.assertEqual(multiline, 3)

    def test_result(self):
        '''"log" attribute of the result object'''
        res = self.recipe(self.raw_frame)
        # Check that we get a not-empty list back
        self.assertTrue(isinstance(res.log, list))
        self.assertNotEqual(len(res.log), 0)
        self.assertTrue(isinstance(res.log[0], logging.LogRecord))

        # Check that we can read debug messages
        self.assertNotEqual(len(res.log.debug), 0)
        self.assertTrue(isinstance(res.log.debug[0], str))
        # Check that we can read info messages
        self.assertNotEqual(len(res.log.info), 0)
        self.assertTrue(isinstance(res.log.info[0], str))
        # Check that we can read warning messages
        self.assertNotEqual(len(res.log.warning), 0)
        self.assertTrue(isinstance(res.log.warning[0], str))
        # Check that there were no error messages
        self.assertEqual(len(res.log.error), 0)
        try:
            res.THE_PRO_CATG_VALUE.close()
        except:
            pass

    def test_error(self):
        '''"log" attribute of the CplError object'''
        try:
            self.recipe('test.fits')
        except cpl.CplError as r:
            res = r
        # Check that we get a not-empty list back
        self.assertTrue(isinstance(res.log, list))
        self.assertNotEqual(len(res.log), 0)
        self.assertTrue(isinstance(res.log[0], logging.LogRecord))
        # Check that we can read debug messages
        self.assertNotEqual(len(res.log.debug), 0)
        self.assertTrue(isinstance(res.log.debug[0], str))
        # Check that we can read info messages
        self.assertNotEqual(len(res.log.info), 0)
        self.assertTrue(isinstance(res.log.info[0], str))
        # Check that we can read warning messages
        self.assertNotEqual(len(res.log.warning), 0)
        self.assertTrue(isinstance(res.log.warning[0], str))
        # Check that we can read error messages
        self.assertNotEqual(len(res.log.error), 0)
        self.assertTrue(isinstance(res.log.error[0], str))
        # Check that we can convert the error to a string
        self.assertTrue(isinstance(res.__str__(), str))
        # Check that we can iterate over error messages
        for r in res:
            self.assertTrue(isinstance(res, cpl.CplError))

class ProcessingInfo(RecipeTestCase):
    def setUp(self):
        RecipeTestCase.setUp(self)
        '''Parameter storage in the result'''
        self.recipe.param.stropt = 'more'
        self.recipe.param.boolopt = False
        self.recipe.param.intopt = 123
        self.recipe.param.floatopt = -0.25
        self.recipe.param.enumopt = 'third'
        self.recipe.param.rangeopt = 0.125
        self.recipe.calib.FLAT = fits.HDUList([
                fits.PrimaryHDU(numpy.random.randint(0, 65001,
                                                     self.image_size))])
        self.res = self.recipe(self.raw_frame).THE_PRO_CATG_VALUE
        self.pinfo = cpl.dfs.ProcessingInfo(self.res)

    def tearDown(self):
        RecipeTestCase.tearDown(self)
        try:
            self.res.close()
        except:
            pass

    def test_list(self):
        '''All processing infos as a list'''
        pi = cpl.dfs.ProcessingInfo.list(self.res[0])
        self.assertTrue(len(pi), 1)
        self.assertTrue(pi[0], self.pinfo)

    def test_param(self):
        '''Parameter information'''
        self.assertEqual(len(self.pinfo.param), len(self.recipe.param))
        for p in self.recipe.param:
            self.assertEqual(self.pinfo.param[p.name], 
                             p.value if p.value is not None else p.default)

    def test_calib(self):
        '''Calibration frame information'''
        self.assertEqual(len(self.pinfo.calib), 1)
        self.assertEqual(self.pinfo.calib['FLAT'][-5:], '.fits')

    def test_tag(self):
        '''Input tag information'''
        self.assertEqual(self.pinfo.tag, self.recipe.tag)

    def test_raw(self):
        '''Raw file information'''
        self.assertEqual(self.pinfo.raw[-5:], '.fits')

    def test_name(self):
        '''Recipe and pipeline name information'''
        self.assertEqual(self.pinfo.name, self.recipe.__name__)
        self.assertEqual(self.pinfo.pipeline, 'iiinstrument')

    def test_version(self):
        '''Version information'''
        self.assertEqual(self.pinfo.version[0], self.recipe.version[0])
        self.assertEqual(self.pinfo.cpl_version, 'cpl-%s' % self.recipe.cpl_version)

    def test_md5(self):
        '''MD5 checksums'''
        md5sum = self.res[0].header.get('DATAMD5')
        self.assertEqual(md5sum, self.pinfo.md5sum)
        md5sum = self.res[0].header.get('HIERARCH ESO PRO REC1 CAL1 DATAMD5')
        self.assertEqual(md5sum, self.pinfo.md5sums[self.pinfo.calib['FLAT']])

    def test_recipe(self):
        '''Recreate and configure the recipe'''
        recipe = self.pinfo.create_recipe()
        self.assertTrue(isinstance(recipe, cpl.Recipe))
        self.assertEqual(len(recipe.calib), 1)
        self.assertEqual(recipe.calib.FLAT.frames[-5:], '.fits')
        self.assertEqual(recipe.__name__, self.recipe.__name__)
        self.assertEqual(recipe.version[0], self.recipe.version[0])
        self.assertEqual(len(recipe.param), len(self.recipe.param))

if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import
import datetime
import logging
import os
import re
import tempfile
import threading

class NullHandler(logging.Handler):
    def emit(self, record):
        pass

logging.getLogger('cpl').addHandler(NullHandler())

level = { "DEBUG":logging.DEBUG, "INFO":logging.INFO, "WARNING":logging.WARN, 
          "ERROR":logging.ERROR, "OFF":(logging.CRITICAL + 1)}

cpl_verbosity = [ logging.DEBUG, logging.INFO, logging.WARN,
                  logging.ERROR, logging.CRITICAL + 1 ]

class LogServer(threading.Thread):

    def __init__(self, name, level = None):
        threading.Thread.__init__(self)
        self.name = name
        self.logger = logging.getLogger(name)
        self.level = cpl_verbosity.index(level) if level is not None else 0
        self.entries = LogList()
        self.regexp = re.compile('(\\d\\d):(\\d\\d):(\\d\\d)' +
                                 '\\s\\[\\s*(\\w+)\\s*\\]' + 
                                 '\\s(\\w+):' +
                                 '(\\s\\[tid=(\\d+)\\])?' +
                                 '\\s(.+)')
        tmphdl, self.logfile = tempfile.mkstemp(prefix = 'cpl', suffix='.log')
        os.close(tmphdl)
        os.remove(self.logfile)
        os.mkfifo(self.logfile)
        self.start()

    def run(self):
        try:
            with open(self.logfile, 'rb', buffering = 0) as logfile:
                line = logfile.readline()
                os.remove(self.logfile)
                while line:
                    self.log(str(line.decode('ascii')))
                    line = logfile.readline()
        except:
            pass

    def log(self, s):
        '''Convert CPL log messages into python log records.
        
        A typical CPL log message looks like

         10:35:25 [WARNING] rtest: [tid=000] No file tagged with FLAT

        '''
        try:
            m = self.regexp.match(s)
            if m is not None:
                g = m.groups()
                creation_date = datetime.datetime.combine(
                    datetime.date.today(), 
                    datetime.time(int(g[0]),int(g[1]),int(g[2])))
                lvl = level.get(g[3], logging.NOTSET)
                func = g[4]
                log = logging.getLogger('%s.%s' % (self.logger.name, func))
                threadid = int(g[6]) if g[6] else None
                msg = g[-1]
                record = logging.LogRecord(log.name, lvl, None, None, 
                                           msg, None, None, func)
                created = float(creation_date.strftime('%s'))
                if record.created < created:
                    created -= 86400
                record.relativeCreated -= record.msecs
                record.relativeCreated += 1000*(created - record.created + 1) 
                record.created = created
                record.msecs = 0.0
                record.threadid = threadid
                record.threadName = ('Cpl-%03i' % threadid) if threadid \
                    else 'CplThread'
            elif self.entries:
                r0 = self.entries[-1]
                msg = s.rstrip()
                lvl = r0.levelno
                log = logging.getLogger(r0.name)
                record = logging.LogRecord(r0.name, lvl, None, None, 
                                           msg, None, None, r0.funcName)
                record.relativeCreated = r0.relativeCreated
                record.created = r0.created
                record.msecs = r0.msecs
                record.threadid = r0.threadid
                record.threadName = r0.threadName
            else:
                return
            self.entries.append(record)
            if log.isEnabledFor(lvl) and log.filter(record):
                log.handle(record)
        except:
            pass

class LogList(list):
    '''List of log messages.

    Accessing this :class:`list` directly will return the
    :class:`logging.LogRecord` instances. 

    Example::

      res = muse_bias(bias_frames)
      for logrecord in res.log:
          print '%s: %s' % (entry.funcname, entry.msg)

    To get them formatted as string, use the :attr:`error`, :attr:`warning`,
    :attr:`info` or :attr:`debug` attributes::

      res = muse_bias(bias_frames)
      for line in res.log.info:
          print line

    '''
    def filter(self, level):
        return [ '%s: %s' % (entry.funcName, entry.msg) for entry in self 
                 if entry.levelno >= level ]

    @property
    def error(self):
        '''Error messages as list of :class:`str`
        '''
        return self.filter(logging.ERROR)

    @property
    def warning(self):
        '''Warnings and error messages as list of :class:`str`
        '''
        return self.filter(logging.WARN)

    @property
    def info(self):
        '''Info, warning and error messages as list of :class:`str`
        '''
        return self.filter(logging.INFO)

    @property
    def debug(self):
        '''Debug, info, warning, and error messages as list of :class:`str`
        '''
        return self.filter(logging.DEBUG)




import textwrap

class Parameter(object):
    '''Runtime configuration parameter of a recipe. 
    Parameters are designed to handle monitor/control data and they provide a
    standard way to pass information to the recipe.

    The CPL implementation supports three classes of parameters: a plain
    value, a value within a given range, or a value as part of an
    enumeration. When a parameter is created it is created for a particular
    value type. In the latter two cases, validation is performed whenever the
    value is set.

    Attributes:

    .. attribute:: Parameter.value

       The value of the parameter, or :obj:`None` if set to default

    .. attribute:: Parameter.default

       The default value of the parameter (readonly).

    .. attribute:: Parameter.name

      The parameter name (readonly). Parameter names are unique. They
      define the identity of a given parameter.

    .. attribute:: Parameter.context

      The parameter context (readonly). The context usually consists of the
      instrument name and the recipe name, separated by a dot. The context is
      used to associate parameters together.

    .. attribute: Parameter.fullname
      The parameter name including the context (readonly).
      The fullname usually consists of the parameter context and the parameter
      name, separated by a dot.

    .. attribute:: Parameter.range 

      The numeric range of a parameter, or :obj:`None` if the parameter range
      is unlimited (readonly).

    .. attribute:: Parameter.sequence

      A :class:`list` of possible values for the parameter if the parameter
      are limited to an enumeration of possible values (readonly).

    The following example prints the attributes of one parameter:

    >>> print 'name:    ', muse_scibasic.param.cr.name
    name:     cr
    >>> print 'fullname:', muse_scibasic.param.cr.fullname
    fullname: muse.muse_scibasic.cr
    >>> print 'context: ', muse_scibasic.param.cr.context
    context:  muse.muse_scibasic
    >>> print 'sequence:', muse_scibasic.param.cr.sequence
    sequence: ['dcr', 'none']
    >>> print 'range:   ', muse_scibasic.param.cr.range
    range:    None
    >>> print 'default: ', muse_scibasic.param.cr.default
    default:  dcr
    >>> print 'value:   ', muse_scibasic.param.cr.value
    value:    None
    '''
    def __init__(self, name):
        self.name = name
        self._value = None

    def _set_attributes(self, context = None, fullname = None, default = None, 
                        desc = None, range_ = None, sequence = None, 
                        ptype = None, enabled = None):
        self.context = context
        self.range = range_
        self.sequence = sequence
        self.default = default
        self.fullname = fullname
        self.type = ptype or default.__class__
        self.enabled = enabled
        self.__doc__ = textwrap.fill("%s (%s; default: %s)" %
                                     (desc, self.type.__name__,
                                      repr(self.default)))

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if value is not None and self.type is not None.__class__:
            if self.type is bool and isinstance(value, str):
                d = {'true':True, 'false':False, 'yes':True, 'no':False}
                value = d.get(value.lower(), value)
            value = self.type(value) 
            if self.sequence and value not in self.sequence:
                raise ValueError("'%s' is not in %s" % (value, self.sequence))
            if self.range and not (self.range[0] <= value <= self.range[-1]):
                raise ValueError("'%s' is not in range %s" % (value, self.range))
        self._value = value

    @value.deleter
    def value(self):
        self._value = None

    def __str__(self):
        return '%s%s' % (
            repr(self.value if self.value is not None else self.default),
            ' (default)' if self.value is None else '')

    def __repr__(self):
        return 'Parameter(%s, %s=%s)' % (
            repr(self.name),
            "value" if self.value is not None else "default",
            repr(self.value if self.value is not None else self.default))

    def __getitem__(self,i):
        return (self.name, self.value or self.default)[i]


class ParameterList(object):
    def __init__(self, recipe, other = None, prefix = None):
        self._recipe = recipe
        self._dict = dict()
        self._pars = list()
        self._prefix = prefix
        childs = set()
        for name, context, fullname, desc, prange, \
                sequence, deflt, ptype, enabled in recipe._recipe.params():
            if prefix:
                if name.startswith(prefix + '.'):
                    aname = name[len(prefix)+1:]
                else:
                    continue
            else:
                aname = name
            if '.' in aname:
                aname = aname.split('.', 1)[0]
                if prefix:
                    aname = prefix + '.' + aname
                childs.add(aname)
            else:
                par = Parameter(name)
                par._set_attributes(context, fullname, deflt,
                                    desc, prange, sequence, ptype, enabled)
                self._dict[name] = par
                self._dict[fullname] = par
                self._dict[self._paramname(aname)] = par
                self._pars.append(par)
        for name in childs:
            clist = ParameterList(recipe, prefix = name)
            self._dict[name] = clist
            for par in clist._pars:
                self._dict[par.name] = par
                self._dict[par.fullname] = par
                self._pars.append(par)
            aname = self._paramname(name)
            self._dict[aname] = clist
        if other:
            self._set_items(other)
        self.__doc__ = self._doc()

    def _set_items(self, other):
        if isinstance(other, self.__class__):
            l = ((o.name, o.value) for o in other)
        elif isinstance(other, dict):
            l = other.items()
        else:
            l = other
        for o in l:
            self[o[0]] = o[1]

    def _del_items(self):
        for p in self:
            del p.value

    @staticmethod
    def _paramname(s):
        for c in [ '-', ' ' ]:
            if isinstance(c, tuple):
                s = s.replace(c[0], c[1])
            else:
                s = s.replace(c, '_')
        return s

    def __iter__(self):
        return self._pars.__iter__()

    def __getitem__(self, key):
        return self._dict[key]

    def __setitem__(self, key, value):
        p = self[key]
        if isinstance(p, self.__class__):
            if value is not None:
                p._set_items(value)
            else:
                p._del_items()
        else:
            p.value = value

    def __delitem__(self, key):
        p = self[key]
        if isinstance(p, self.__class__):
            p._del_items()
        else:
            del p.value

    def __str__(self):
        return dict(iter(self)).__str__()
    
    def __contains__(self, key):
        return key in self._dict

    def __len__(self):
        return len(self._pars)
        
    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        if key.startswith('_'):
            super(ParameterList, self).__setattr__(key, value)
        else:
            self[key] = value

    def __delattr__(self, key):
        del self[key]

    def __dir__(self):
        return list(set(self._paramname(d) 
                        for d in self._dict.keys() if '.' not in d))

    def __repr__(self):
        return repr(dict(iter(self)))

    def __eq__(self, other):
        return dict(iter(self)) == other

    def _doc(self):
        if len(self) == 0:
            return 'No parameters'
        r = ''
        maxlen = max(len(p.name) for p in self)
        for p in self:
            r += textwrap.fill(
                p.__doc__,
                subsequent_indent = ' ' * (maxlen + 3),
                initial_indent = ' %s: ' % p.name.rjust(maxlen)) + '\n'
        return r        

    def _aslist(self, par):
        parlist = ParameterList(self._recipe, self)
        if par is not None:
            parlist._set_items(par.items())
        l = list()
        for param in parlist:
            if isinstance(param, Parameter):
                if param.value is not None:
                    l.append((param.fullname, param.value))
            else:
                l += param._aslist(par)
        return l



import sys
from astropy.io import fits

import cpl

class ProcessingInfo(object):
    '''Support for reading input files and parameters from the FITS
    header of a CPL processed file.

    This is done through the FITS headers that were written by the DFS function
    called within the processing recipe.

    .. attribute:: name 

       Recipe name

    .. attribute:: version 

       Recipe version string

    .. attribute::  pipeline

       Pipeline name

    .. attribute:: cpl_version
    
       CPL version string

    .. attribute:: tag

       Tag name

    .. attribute:: calib

       Calibration frames from a FITS file processed with CPL.
       The result of this function may directly set as :attr:`cpl.Recipe.calib`
       attribute::
    
         import cpl
         myrecipe = cpl.Recipe('muse_bias')
         myrecipe.calib = cpl.dfs.ProcessingInfo('MASTER_BIAS_0.fits').calib

       .. note::

          This will not work properly for files that had
          :class:`astropy.io.fits.HDUList` inputs since they have assigned a
          temporary file name only.

    .. attribute:: raw

       Raw (input) frames

       .. note::

          This will not work properly for files that had
          :class:`astropy.io.fits.HDUList` inputs since they have assigned a
          temporary file name only.

    .. attribute:: param

       Processing parameters.
       The result of this function may directly set as :attr:`cpl.Recipe.param`
       attribute::
    
         import cpl
         myrecipe = cpl.Recipe('muse_bias')
         myrecipe.param = cpl.dfs.ProcessingInfo('MASTER_BIAS_0.fits').param

    .. attribute:: md5sum

       MD5 sum of the data portions of the output file (header keyword 
       'DATAMD5').

    .. attribute:: md5sums

       MD5 sums of the input and calibration files. :class:`dict` with the
       file name as key and the corresponding MD5 sum as value.

       .. note::

          Due to a design decision in CPL, the raw input files are not
          accompanied with the MD5 sum.
    '''

    def __init__(self, source, recno = -1):
        '''
        :param source: Object pointing to the result file header
        :type source: :class:`str` or :class:`astropy.io.fits.HDUList`
                      or :class:`astropy.io.fits.PrimaryHDU` or 
                      :class:`astropy.io.fits.Header`
        :param recno: Record number. Optional. If not given, the last record
                      (with the highest record number) is used.
        :type recno: :class:`int`
        '''
        if isinstance(source, str):
            header = fits.open(source)[0].header
        elif isinstance(source, (fits.HDUList, list)):
            header = source[0].header
        elif isinstance(source, fits.PrimaryHDU):
            header = source.header
        elif isinstance(source, (fits.Header, dict)):
            header = source
        else:
            raise ValueError('Cannot assign type {0} to header'.format(
                    source.__class__.__name__))
        
        if recno < 0:
            for reccnt in range(1, 2**16):
                if 'HIERARCH ESO PRO REC{0} ID'.format(reccnt) not in header:
                    break
            recno += reccnt
        self.name = header['HIERARCH ESO PRO REC{0} ID'.format(recno)]
        self.product = header['HIERARCH ESO PRO CATG']
        self.orig_filename = header['PIPEFILE']
        pipe_id = header.get('HIERARCH ESO PRO REC{0} PIPE ID'.format(recno))
        if pipe_id:
            self.pipeline, version = pipe_id.split('/')
            num_version = 0
            for i in version.split('.'):
                num_version = num_version * 100 + int(i)
            self.version = (num_version, version)
        else:
            self.pipeline =  None
            self.version = None
        self.cpl_version = header.get('HIERARCH ESO PRO REC{0} DRS ID'.format(recno))
        self.md5sum = header.get('DATAMD5')
        self.md5sums = {}
        self.calib = ProcessingInfo._get_rec_keys(header, recno, 'CAL', 'CATG', 'NAME')
        for cat, md5 in ProcessingInfo._get_rec_keys(header, recno, 'CAL', 'CATG',
                                                     'DATAMD5').items():
            if isinstance(md5, list):
                for i, m in enumerate(md5):
                    if m is not None:
                        self.md5sums[self.calib[cat][i]] = m
            elif md5 is not None:
                self.md5sums[self.calib[cat]] = md5
        raw = ProcessingInfo._get_rec_keys(header, recno, 'RAW', 'CATG', 'NAME')
        if raw:
            self.tag = list(raw.keys())[0]
            self.raw = raw[self.tag]
            md5 = ProcessingInfo._get_rec_keys(header, recno, 'RAW', 'CATG',
                                               'DATAMD5')[self.tag]
            if isinstance(md5, list):
                for i, m in enumerate(md5):
                    if m is not None: 
                        self.md5sums[self.raw[i]] = m
            elif md5 is not None:
                self.md5sums[self.raw] = md5
        else:
            self.tag = None
            self.raw = None
            self.input = None
        param = ProcessingInfo._get_rec_keys(header, recno, 'PARAM', 'NAME', 'VALUE')
        self.param = dict()
        for k,v in param.items():
            self.param[k] = ProcessingInfo._best_type(v)
            
    def create_recipe(self):
        '''Create a recipe and configure it with the parameters, calibration frames,
        and the input tag. The recipe version will be the latest available one.
        '''
        recipe = cpl.Recipe(self.name)
        recipe.param = self.param
        recipe.calib = self.calib
        recipe.tag = self.tag
        return recipe

    def create_script(self, scriptfile = sys.stdout):
        '''Create a sample script that creates the recipe, configures it with
        the parameters, calibration frames and input tags, and finally
        starts the recipe.
        '''
        if isinstance(scriptfile, str):
            scriptfile = file(scriptfile, mode='w')
        scriptfile.write('import cpl\n\n')
        scriptfile.write('# Recipe: {0}.{1}, Version {2}, CPL version {3}\n'.format(
                self.pipeline, self.name, self.version[1], self.cpl_version))
        scriptfile.write('{0} = cpl.Recipe({1}, version = {2})\n'.format(
                self.name, repr(self.name), repr(self.version[0])))
        scriptfile.write('\n# Parameters:\n')
        for k,v in self.param.items():
            scriptfile.write('{0}.param.{1} = {2}\n'.format(self.name, k, repr(v)))
        if self.calib:
            scriptfile.write('\n# Calibration frames:\n')
        for k,v in self.calib.items():
            scriptfile.write('{0}.calib.{1} = {2}\n'.format(self.name, k, repr(v)))
        scriptfile.write('\n# Process input frames:\n')
        scriptfile.write('{0}.tag = {1}\n'.format(self.name, repr(self.tag)))
        scriptfile.write('res = {0}({1})\n'.format(self.name, repr(self.raw)))
#        scriptfile.write('{0} = res.{1}\n'.format(self.product.lower(), self.product))
#        scriptfile.write('{0}.writeto({1})\n'.format(self.product.lower(),
#                                                     repr(self.orig_filename)))

    def __str__(self):
        s = 'Recipe: {0}, Version {1}, CPL version {2}\n'.format(
            self.name, self.version, self.cpl_version)
        s += 'Parameters:\n'
        for k,v in self.param.items():
            s += ' {0}.{1}.{2} = {3}\n'.format(self.pipeline, self.name, k, v)
        if self.calib:
            s += 'Calibration frames:\n'
        for k,v in self.calib.items():
            if isinstance(v, (str, unicode)):
                s += ' {0} {1}\n'.format(v,k)
            else:
                m = max(len(n) for n in v)
                for n in v:
                    s += ' {0:<{width}} {1}\n'.format(n, m, width = m)
        if self.raw is not None:
            s += 'Input frames:\n'
            if isinstance(self.raw, (str, unicode)):
                s += ' {0} {1}\n'.format(self.raw, self.tag)
            else:
                m = max(len(n) for n in self.raw)
                for n in self.raw:
                    s += ' {0:<{width}} {1}\n'.format(n, self.tag, width = m)
        return s

    def printinfo(self):
        '''Print the recipe information to standard output.
        '''
        print(str(self))

    @staticmethod
    def _get_rec_keys(header, recno, key, name, value):
        '''Get a dictionary of key/value pairs from the DFS section of the
        header.

        :param key: Common keyword for the value. Usually 'PARAM' for
                    parameters, 'RAW' for raw frames, and 'CAL' for
                    calibration frames.
        :type key: :class:`str`
        :param recno: Record number.
        :type recno: :class:`int`
        :param name: Header keyword (last part) for the name of each key
        :type name: :class:`str`
        :param value: Header keyword (last part) for the value of each key
        :type name: :class:`str`

        When the header

          HIERARCH ESO PRO REC1 PARAM1 NAME = 'nifu'
          HIERARCH ESO PRO REC1 PARAM1 VALUE = '1'
          HIERARCH ESO PRO REC1 PARAM2 NAME = 'combine'
          HIERARCH ESO PRO REC1 PARAM2 VALUE = 'median'

        is called with

          ProcessingInfo._get_rec_keys(1, 'PARAM', 'NAME', 'VALUE')

        the returned dictionary will contain the keys

          res['nifu'] = '1'
          res['combine'] = 'median'
        '''
        res = dict()
        for i in range(1, 2**16):
            try:
                prefix = 'HIERARCH ESO PRO REC{0} {1}{2}'.format(recno, key, i)
                k = header['{0} {1}'.format(prefix, name)]
                fn = header.get('{0} {1}'.format(prefix, value))
                if k not in  res:
                    res[k] = fn
                elif isinstance(res[k], list):
                    res[k].append(fn)
                else:
                    res[k] = [ res[k], fn ]
            except KeyError:
                break
        return res

    @staticmethod
    def _best_type(value):
        '''Convert the value to the best applicable type: :class:`int`,
        :class:`float`, :class:`bool` or :class`str`.

        :param value: Value to convert.
        :type value: :class:`str`
        '''
        for t in int, float:
            try:
                return t(value)
            except ValueError:
                pass
        return {'true':True, 'false':False}.get(value, value)

    @staticmethod
    def list(source):
        '''Get a list of all `ProcessingInfo` objects in the FITS header. The
        list is sorted by the execution order.

        :param source: Object pointing to the result file header
        :type source: :class:`str` or :class:`astropy.io.fits.HDUList`
                      or :class:`astropy.io.fits.PrimaryHDU` or
                      :class:`astropy.io.fits.Header`
        '''
        pi = []
        for i in range(1, 2**16):
            try:
                pi.append(ProcessingInfo(source, i))
            except KeyError:
                break
        return pi

if __name__ == '__main__':
    for arg in sys.argv[1:]:
        print('{0}\nfile: {1}'.format('-' * 72, arg))
        pi = cpl.dfs.ProcessingInfo(arg)
        pi.printinfo()



import os
from setuptools import setup, Extension

author = 'Ole Streicher'
email = 'python-cpl@liska.ath.cx'
license_ = 'GPL'
cpl_version = '0.7.4'
description = "Python interface for the ESO Common Pipeline Library"
long_description = '''\
This module can list, configure and execute CPL-based recipes from Python
(python2 and python3).  The input, calibration and output data can be
specified as FITS files or as ``astropy.io.fits`` objects in memory.

The ESO `Common Pipeline Library <http://www.eso.org/sci/software/cpl/>`_
(CPL) comprises a set of ISO-C libraries that provide a comprehensive,
efficient and robust software toolkit. It forms a basis for the creation of
automated astronomical data-reduction tasks. One of the features provided by
the CPL is the ability to create data-reduction algorithms that run as plugins
(dynamic libraries). These are called "recipes" and are one of the main
aspects of the CPL data-reduction development environment.
'''

pkgname = 'python-cpl'
baseurl = ('https://files.pythonhosted.org/packages/source/'
           '{0}/{1}/{1}-{2}.tar.gz'.format(pkgname[0], pkgname, cpl_version))
classifiers = '''Development Status :: 4 - Beta
Intended Audience :: Science/Research
License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)
Operating System :: MacOS :: MacOS X
Operating System :: POSIX
Operating System :: Unix
Programming Language :: C
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 3
Topic :: Scientific/Engineering :: Astronomy
'''.splitlines()


def create_version_file(cpl_version=cpl_version):
    with open(os.path.join('cpl', 'version.py'), 'w') as vfile:
        vfile.write("version = %s\n" % repr(cpl_version))
        vfile.write("author = %s\n" % repr(author))
        vfile.write("email = %s\n" % repr(email))
        vfile.write("license_ = %s\n" % repr(license_))


try:
    create_version_file()
except IOError:
    pass

module1 = Extension('cpl.CPL_recipe',
                    sources=['cpl/CPL_recipe.c', 'cpl/CPL_library.c'])

setup(
    name=pkgname,
    version=cpl_version,
    author=author,
    author_email=email,
    description=description,
    long_description=long_description,
    license=license_,
    url='https://pypi.org/project/%s/%s' % (pkgname, cpl_version),
    download_url='%s/%s-%s.tar.gz' % (baseurl, pkgname, cpl_version),
    classifiers=classifiers,
    python_requires='>=2.7',
    install_requires=['astropy'],
    provides=['cpl'],
    packages=['cpl'],
    ext_modules=[module1]
)



'''Python interface for the Common Pipeline Library

This module can list, configure and execute CPL-based recipes from Python
(python2 and python3).  The input, calibration and output data can be
specified as FITS files or as ``astropy.io.fits`` objects in memory.

The ESO `Common Pipeline Library <http://www.eso.org/sci/software/cpl/>`_
(CPL) comprises a set of ISO-C libraries that provide a comprehensive,
efficient and robust software toolkit. It forms a basis for the creation of
automated astronomical data-reduction tasks. One of the features provided by
the CPL is the ability to create data-reduction algorithms that run as plugins
(dynamic libraries). These are called "recipes" and are one of the main
aspects of the CPL data-reduction development environment.
'''

from __future__ import absolute_import

from .version import version as __version__
from .version import author as __author__
from .version import email as __email__
from .version import license_ as __license__

from .recipe import Recipe
from .param import Parameter
from .frames import FrameConfig
from .result import Result, CplError, RecipeCrash
from . import dfs
from . import esorex
from . import CPL_recipe

Recipe.dir = '.'
cpl_versions = [ '%i.%i.%i' % ver for ver in CPL_recipe.cpl_versions() ]
del CPL_recipe
del absolute_import
del recipe, version, param, frames, result, md5sum
try:
    del ver
except NameError:
    pass
